import json
import logging
from dataclasses import field
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import marshmallow_dataclass
from marshmallow import ValidationError

from ggshield.core import ui
from ggshield.core.config.utils import (
    find_global_config_path,
    find_local_config_path,
    load_yaml_dict,
    remove_common_dict_items,
    replace_dash_in_keys,
    save_yaml_dict,
    update_dict_from_other,
)
from ggshield.core.config.v1_config import convert_v1_config_dict
from ggshield.core.constants import DEFAULT_LOCAL_CONFIG_PATH
from ggshield.core.errors import ParseError, UnexpectedError, format_validation_error
from ggshield.core.types import FilteredConfig, IgnoredMatch


logger = logging.getLogger(__name__)
CURRENT_CONFIG_VERSION = 2

_IGNORE_KNOWN_SECRETS_KEY = "ignore_known_secrets"


@marshmallow_dataclass.dataclass
class SecretConfig(FilteredConfig):
    """
    Holds all user-defined secret-specific settings
    """

    show_secrets: bool = False
    ignored_detectors: Set[str] = field(default_factory=set)
    ignored_matches: List[IgnoredMatch] = field(default_factory=list)
    ignored_paths: Set[str] = field(default_factory=set)
    ignore_known_secrets: bool = False
    with_incident_details: bool = False
    # if configuration key is left unset the dashboard's remediation message is used.
    all_secrets: bool = False
    prereceive_remediation_message: str = ""

    def add_ignored_match(self, secret: IgnoredMatch) -> None:
        """
        Add secret to ignored_matches.
        """
        for match in self.ignored_matches:
            if match.match == secret.match:
                # take the opportunity to name the ignored match
                if not match.name:
                    match.name = secret.name
                return
        self.ignored_matches.append(secret)

    def dump_for_monitoring(self) -> str:
        return json.dumps(
            {
                "show_secrets": self.show_secrets,
                "ignored_detectors_count": len(self.ignored_detectors),
                "ignored_matches_count": len(self.ignored_matches),
                "ignored_paths_count": len(self.ignored_paths),
                "ignore_known_secrets": self.ignore_known_secrets,
                "with_incident_details": self.with_incident_details,
                "has_prereceive_remediation_message": bool(
                    self.prereceive_remediation_message
                ),
                "all_secrets": self.all_secrets,
            }
        )


@marshmallow_dataclass.dataclass
class UserConfig(FilteredConfig):
    """
    Holds all ggshield settings defined by the user in the .gitguardian.yaml files
    (local and global).
    """

    instance: Optional[str] = None
    exit_zero: bool = False
    verbose: bool = False
    allow_self_signed: bool = False
    max_commits_for_hook: int = 50
    secret: SecretConfig = field(default_factory=SecretConfig)
    debug: bool = False

    # If we hit any deprecated syntax when loading a configuration file, we do not
    # display them directly, otherwise the messages would also be shown when running
    # `ggshield config migrate`, which would be odd.
    # Instead, we store them in this list and a result_callback() function displays
    # them when we quit. When `config migrate` runs, it clears this list, so nothing
    # gets displayed.
    deprecation_messages: List[str] = field(default_factory=list)

    def save(self, config_path: Path) -> None:
        """
        Save config to config_path
        """
        dct = self.to_config_dict()
        save_yaml_dict(dct, config_path)

    def to_config_dict(self) -> Dict[str, Any]:
        dct = self.to_dict()
        default_dct = UserConfig.from_dict({}).to_dict()

        dct = remove_common_dict_items(dct, default_dct)

        dct["version"] = CURRENT_CONFIG_VERSION

        return dct

    @classmethod
    def load(cls, config_path: Optional[Path] = None) -> Tuple["UserConfig", Path]:
        """
        Load the various user configs files to create a UserConfig object:
        - global user configuration file (in the home)
        - local user configuration file (in the repository)

        Returns a UserConfig instance, and the path where updates should be saved
        """
        deprecation_messages: List[str] = []
        if config_path:
            logger.debug("Loading custom config from %s", config_path)
            dct = _load_config_dict(config_path, deprecation_messages)
            user_config = UserConfig.from_config_dict(dct)
            user_config.deprecation_messages = deprecation_messages
            return user_config, config_path

        user_config_dict: Dict[str, Any] = {}
        global_config_path = find_global_config_path()
        if global_config_path:
            dct = _load_config_dict(global_config_path, deprecation_messages)
            update_dict_from_other(user_config_dict, dct)
            logger.debug("Loaded global config from %s", global_config_path)
        else:
            logger.debug("No global config")

        local_config_path = find_local_config_path()
        if local_config_path:
            dct = _load_config_dict(local_config_path, deprecation_messages)
            update_dict_from_other(user_config_dict, dct)
            config_path = local_config_path
            logger.debug("Loaded local config from %s", local_config_path)
        else:
            logger.debug("No local config")

        user_config = UserConfig.from_config_dict(user_config_dict)
        user_config.deprecation_messages = deprecation_messages

        if config_path is None:
            config_path = Path(DEFAULT_LOCAL_CONFIG_PATH)
        return user_config, config_path

    @staticmethod
    def from_config_dict(data: Dict[str, Any]) -> "UserConfig":
        """Create a UserConfig instance. In case of error, format it and raise
        ParseError."""
        try:
            return UserConfig.from_dict(data)
        except ValidationError as exc:
            message = format_validation_error(exc)
            raise ParseError(message) from exc


UserConfig.SCHEMA = marshmallow_dataclass.class_schema(UserConfig)(
    exclude=("deprecation_messages",)
)


def _load_config_dict(
    config_path: Path, deprecation_messages: List[str]
) -> Dict[str, Any]:
    """
    Load configuration from `config_path` as a dict.
    Appends any deprecation message regarding this file to `deprecation_messages`.
    """
    try:
        # load_yaml_dict() returns None if `config_path` does not exist. When this
        # happens, initialize the config dict to an empty configuration file,
        # but set the version to the latest version so that we don't get deprecation
        # messages.
        dct = load_yaml_dict(config_path) or {"version": CURRENT_CONFIG_VERSION}
    except ValueError as exc:
        raise ParseError(str(exc)) from exc

    dash_keys = replace_dash_in_keys(dct)

    config_version = dct.pop("version", 1)
    if config_version == 2:
        if dash_keys:
            _warn_about_dash_keys(config_path, dash_keys)
        _fix_ignore_known_secrets(dct)
    elif config_version == 1:
        deprecation_messages.append(
            f"{config_path} uses a deprecated configuration file format."
            " Run `ggshield config migrate` to migrate it to the latest version."
        )
        dct = convert_v1_config_dict(dct, deprecation_messages)
    else:
        raise UnexpectedError(f"Don't know how to load config version {config_version}")
    return dct


def _fix_ignore_known_secrets(data: Dict[str, Any]) -> None:
    """Fix a mistake done when implementing ignore-known-secrets: since this is a secret
    specific key, it should have been stored in the "secret" mapping, not in the root
    one"""

    value = data.pop(_IGNORE_KNOWN_SECRETS_KEY, None)
    if value is None:
        # No key to fix
        return

    secret_dct = data.setdefault("secret", {})
    if _IGNORE_KNOWN_SECRETS_KEY in secret_dct:
        # Do not override if it's already there
        return
    secret_dct[_IGNORE_KNOWN_SECRETS_KEY] = value


def _warn_about_dash_keys(config_path: Path, dash_keys: Set[str]) -> None:
    for old_key in sorted(dash_keys):
        new_key = old_key.replace("-", "_")
        ui.display_warning(
            f"{config_path}: Config key {old_key} is deprecated, use {new_key} instead."
        )



import os
from enum import Enum

from ggshield.core.errors import UnexpectedError


class SupportedCI(Enum):
    GITLAB = "GITLAB"
    TRAVIS = "TRAVIS"
    CIRCLECI = "CIRCLECI"
    JENKINS = "JENKINS HOME"
    GITHUB = "GITHUB ACTIONS"
    BITBUCKET = "BITBUCKET PIPELINES"
    DRONE = "DRONE"
    AZURE = "AZURE PIPELINES"

    @staticmethod
    def from_ci_env() -> "SupportedCI":
        if os.getenv("GITLAB_CI"):
            return SupportedCI.GITLAB
        if os.getenv("GITHUB_ACTIONS"):
            return SupportedCI.GITHUB
        if os.getenv("TRAVIS"):
            return SupportedCI.TRAVIS
        if os.getenv("JENKINS_HOME") or os.getenv("JENKINS_URL"):
            return SupportedCI.JENKINS
        if os.getenv("CIRCLECI"):
            return SupportedCI.CIRCLECI
        if os.getenv("BITBUCKET_COMMIT"):
            return SupportedCI.BITBUCKET
        if os.getenv("DRONE"):
            return SupportedCI.DRONE
        if os.getenv("BUILD_BUILDID"):
            return SupportedCI.AZURE

        raise UnexpectedError(
            f"Current CI is not detected or supported."
            f" Supported CIs: {', '.join([ci.value for ci in SupportedCI])}."
        )



import re
from typing import Any, List, Optional, Tuple

import click
import requests

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client
from ggshield.core.config import Config
from ggshield.core.constants import DEFAULT_INSTANCE_URL
from ggshield.core.errors import UnexpectedError
from ggshield.core.url_utils import clean_url
from ggshield.verticals.auth import OAuthClient


def validate_login_path(
    instance: Optional[str], sso_url: Optional[str]
) -> Tuple[Optional[str], Optional[str]]:
    """
    Validate that the SSO URL and the instance refer to the same instance if they are both defined,
    that the SSO URL has a correct format
    and return the couple (instance, login_path) that will be used to redirect to the login page
    """
    if sso_url is None:
        return instance, None
    sso_parsed_url = clean_url(sso_url)
    if (
        not sso_parsed_url.scheme
        or not sso_parsed_url.netloc
        or not re.match(
            r"/auth/sso/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
            sso_parsed_url.path,
        )
    ):
        raise click.BadParameter(
            "Please provide a valid SSO URL.",
            param_hint="sso-url",
        )
    sso_instance = f"{sso_parsed_url.scheme}://{sso_parsed_url.netloc}"
    sso_login_path = sso_parsed_url.path

    if instance is None:
        return sso_instance, sso_login_path

    config_parsed_url = clean_url(instance)
    if (
        config_parsed_url.scheme != sso_parsed_url.scheme
        or config_parsed_url.netloc != sso_parsed_url.netloc
    ):
        raise UnexpectedError("instance and SSO URL params do not match")
    return instance, sso_login_path


def print_default_instance_message(config: Config) -> None:
    """If the instance used is not defined as the default instance, show a message
    explaining how to make it the default instance."""
    cli_instance = config.cmdline_instance_name
    if not cli_instance or cli_instance == DEFAULT_INSTANCE_URL:
        return
    click.echo(
        "\nTo make ggshield always use this instance,"
        f' run "ggshield config set instance {cli_instance}".'
    )


@click.command()
@click.option(
    "--method",
    required=False,
    default="web",
    type=click.Choice(["token", "web"]),
    help="Authentication method.",
)
@click.option(
    "--instance",
    required=False,
    type=str,
    help="URL of the instance to authenticate against.",
    metavar="URL",
)
@click.option(
    "--scopes",
    required=False,
    type=str,
    help=(
        "Space-separated list of extra scopes to request in addition to the default"
        " `scan` scope."
    ),
    metavar="SCOPES",
)
@click.option(
    "--sso-url",
    required=False,
    type=str,
    help="URL of your SSO login page to force the authentication flow through your workspace SSO.",
    metavar="URL",
)
@click.option(
    "--token-name",
    required=False,
    type=str,
    help="Name of new token.",
)
@click.option(
    "--lifetime",
    required=False,
    type=click.IntRange(0),
    default=None,
    help="Number of days before the token expires. 0 means the token never expires.",
    metavar="DAYS",
)
@add_common_options()
@click.pass_context
def login_cmd(
    ctx: click.Context,
    method: str,
    instance: Optional[str],
    scopes: Optional[str],
    token_name: Optional[str],
    lifetime: Optional[int],
    sso_url: Optional[str],
    **kwargs: Any,
) -> int:
    """
    Authenticate with a GitGuardian instance.

    A successful authentication results in a personal access token.
    This token is stored in your configuration and used to authenticate your future requests.

    The default authentication method is `web`.
    ggshield launches a web browser to authenticate you to your GitGuardian instance,
    then automatically generates a token on your behalf.

    Alternatively, you can use `--method token` to authenticate using an already existing token.
    The minimum required scope for the token is `scan`.

    By default, the created token will have the `scan` scope. Use the `--scopes` option
    to grant the token extra scopes. You can find the list of available scopes in
    [GitGuardian API documentation][1].

    If a valid personal access token is already configured, this command simply displays
    a success message indicating that ggshield is already ready to use.

    [1]: https://docs.gitguardian.com/api-docs/introduction#scopes
    """
    config = ContextObj.get(ctx).config

    if method != "web":
        if sso_url is not None:
            raise click.BadParameter(
                "--sso-url is reserved for the web login method.", param_hint="sso-url"
            )

        if scopes is not None:
            raise click.BadParameter(
                "--scopes is reserved for the web login method.", param_hint="scopes"
            )

    if method == "token":
        token_login(config, instance)
        return 0

    if method == "web":
        extra_scopes = scopes.split(" ") if scopes else None
        web_login(config, instance, token_name, lifetime, sso_url, extra_scopes)
        return 0

    return 1


def token_login(config: Config, instance: Optional[str]) -> None:
    if instance:
        config.cmdline_instance_name = instance
    instance = config.instance_name
    # Override instance to make sure we get a normalized instance name
    instance_config = config.auth_config.get_or_create_instance(instance_name=instance)

    token = None
    if not click.get_text_stream("stdin").isatty():
        # Read from stdin only when the stdin is not connected to terminal, but is provided from the pipe
        token = click.get_text_stream("stdin").read().strip()

    # Prompt if the token was not provided with stdin
    if not token:
        token = click.prompt("Enter your GitGuardian API token", hide_input=True)

    if not token:
        raise UnexpectedError("No API token was provided.")

    # enforce using the token (and not use config default)
    client = create_client(
        api_key=token,
        api_url=config.api_url,
        allow_self_signed=config.user_config.allow_self_signed,
    )
    try:
        response = client.get(endpoint="token")
    except requests.exceptions.ConnectionError as e:
        if "Failed to resolve" in str(e):
            raise click.UsageError(f"Invalid instance: {instance}.")
        else:
            raise UnexpectedError(f"Failed to connect to {instance}.") from e

    if not response.ok:
        raise UnexpectedError("Authentication failed with token.")

    api_token_data = response.json()
    scopes = api_token_data["scope"]
    if "scan" not in scopes:
        raise UnexpectedError("This token does not have the scan scope.")

    instance_config.init_account(token, api_token_data)
    config.auth_config.save()
    click.echo("Authentication was successful.")
    print_default_instance_message(config)


def web_login(
    config: Config,
    instance: Optional[str],
    token_name: Optional[str],
    lifetime: Optional[int],
    sso_url: Optional[str],
    extra_scopes: Optional[List[str]],
) -> None:
    instance, login_path = validate_login_path(instance=instance, sso_url=sso_url)
    if instance:
        config.cmdline_instance_name = instance
    defined_instance = config.instance_name
    # Override instance to make sure we get a normalized instance name
    config.auth_config.get_or_create_instance(instance_name=defined_instance)

    client = OAuthClient(config, defined_instance)

    if client.check_existing_token():
        # skip the process if a valid token is already saved
        return

    client.oauth_process(
        token_name=token_name,
        lifetime=lifetime,
        login_path=login_path,
        extra_scopes=extra_scopes,
    )
    print_default_instance_message(config)



import time

from pygitguardian import GGClientCallbacks

from ggshield.core import ui


RATE_LIMIT_MESSAGE_MINIMUM_INTERVAL = 2


class ClientCallbacks(GGClientCallbacks):
    """Implementation of GGClientCallbacks using GGShieldUI to show messages"""

    def __init__(self):
        self._last_rate_limit_message_at = 0.0

    def on_rate_limited(self, delay: int) -> None:
        # When we are rate-limited, all scanning threads report the rate limit. To avoid
        # spamming the console, only show a message if the previous rate limit message
        # was more than RATE_LIMIT_MESSAGE_MINIMUM_INTERVAL seconds ago
        now = time.time()
        if now - self._last_rate_limit_message_at < RATE_LIMIT_MESSAGE_MINIMUM_INTERVAL:
            return
        ui.display_warning(f"Rate-limit hit, retrying in {delay} seconds")
        self._last_rate_limit_message_at = now



import logging
import os
import platform
import sys
from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path
from typing import Iterator, Optional, Tuple, Union, overload


logger = logging.getLogger(__name__)


@lru_cache(None)
def get_os_info() -> Tuple[str, str]:
    """
    Returns a tuple of [OS, version]

    OS is always in lowercase
    """
    if sys.platform.lower() == "linux":
        return parse_os_release(Path("/etc/os-release"))
    else:
        return platform.system().lower(), platform.version()


def parse_os_release(os_release_path: Path) -> Tuple[str, str]:
    """
    Extract and return Linux's OS-name and OS-Version
    If extraction fails, we return ('linux', 'unknown')
    """
    error_tuple = "linux", "unknown"

    try:
        with open(os_release_path) as f:
            lines = f.readlines()

        # Build a dictionary from the os-release file contents
        data_dict = {}
        for line in lines:
            if "=" in line:
                key, value = line.split("=")
                key, value = key.strip(), value.strip()
                data_dict[key] = value.strip('"')

        if "ID" not in data_dict:
            return error_tuple

        return data_dict["ID"], data_dict.get("VERSION_ID", "unknown")
    except Exception as exc:
        logger.warning(f"Failed to read Linux OS name and version: {exc}")
        return error_tuple


@contextmanager
def cd(newdir: Union[str, Path]) -> Iterator[None]:
    """
    A context manager to temporarily change the current directory
    """
    prevdir = Path.cwd()
    newdir = Path(newdir).expanduser()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)


@overload
def getenv_int(key: str, default: None = None) -> Optional[int]: ...


@overload
def getenv_int(key: str, default: int) -> int: ...


def getenv_int(key: str, default: Optional[int] = None) -> Optional[int]:
    value = os.getenv(key)
    if value is None:
        return default
    return int(value)


@overload
def getenv_float(key: str, default: None = None) -> Optional[float]: ...


@overload
def getenv_float(key: str, default: float) -> float: ...


def getenv_float(key: str, default: Optional[float] = None) -> Optional[float]:
    value = os.getenv(key)
    if value is None:
        return default
    return float(value)


@overload
def getenv_bool(key: str, default: None = None) -> Optional[bool]: ...


@overload
def getenv_bool(key: str, default: bool) -> bool: ...


def getenv_bool(key: str, default: Optional[bool] = None) -> Optional[bool]:
    value = os.getenv(key)
    if value is None:
        return default
    return value.lower() not in {"false", "0"}



from pathlib import Path
from unittest.mock import Mock, patch

import pytest
import requests.exceptions
from _pytest.monkeypatch import MonkeyPatch
from pyfakefs.fake_filesystem import FakeFilesystem

import ggshield.core
from ggshield.core.check_updates import (
    CACHE_FILE,
    check_for_updates,
    load_last_check_time,
)


@patch("requests.get")
@pytest.mark.parametrize(
    "current_version,remote_version,expected_latest_version",
    (
        ("1.2.3", "1.2.4", "1.2.4"),
        ("1.2.3", "1.3.3", "1.3.3"),
        ("1.2.3", "2.2.3", "2.2.3"),
        ("1.2.3", "1.2.3", None),
        ("1.2.3", "1.2.2", None),
        ("1.2.3", "1.1.3", None),
        ("1.2.3", "0.2.3", None),
    ),
)
def test_check_for_updates(
    request_get_mock: Mock,
    current_version: str,
    remote_version: str,
    expected_latest_version: bool,
    fs: FakeFilesystem,
    monkeypatch: MonkeyPatch,
):
    """
    GIVEN the latest released version
    WHEN calling check_for_updates
    THEN the latest version is returned if it's newer
    """
    monkeypatch.setattr(ggshield.core.check_updates, "__version__", current_version)
    request_get_mock.return_value.status_code = 200
    request_get_mock.return_value.json.return_value = {"tag_name": f"v{remote_version}"}

    latest_version = check_for_updates()

    assert latest_version == expected_latest_version
    assert fs.exists(CACHE_FILE)


@patch("requests.get")
def test_check_for_updates_twice_only_notifies_once(
    request_get_mock: Mock, fs: FakeFilesystem, monkeypatch: MonkeyPatch
):
    """
    GIVEN a first check_for_updates() call
    WHEN calling check_for_updates() a second time
    THEN no network calls are made and no update is reported
    """
    monkeypatch.setattr(ggshield.core.check_updates, "__version__", "1.0")

    request_get_mock.return_value.status_code = 200
    request_get_mock.return_value.json.return_value = {"tag_name": "v1.1"}
    latest_version = check_for_updates()
    assert latest_version == "1.1"

    request_get_mock.reset_mock()
    latest_version = check_for_updates()

    request_get_mock.assert_not_called()
    assert latest_version is None


@patch("requests.get")
def test_check_for_updates_request_exceptions_are_caught(
    request_get_mock: Mock, fs: FakeFilesystem
):
    """
    GIVEN an environment with no network access to api.github.com
    WHEN check_for_updates() is called
    THEN it does not fail
    AND the check time is recorded to avoid rechecking every time
    """
    request_get_mock.side_effect = requests.exceptions.ConnectTimeout()
    latest_version = check_for_updates()
    assert latest_version is None

    assert fs.exists(CACHE_FILE)


@patch("requests.get")
def test_check_for_updates_does_nothing_if_cache_cant_be_saved(
    request_get_mock: Mock, fs: FakeFilesystem
):
    """
    GIVEN an environment where it is not possible to save the check time
    WHEN check_for_updates() is called
    THEN it does not fail
    AND no network access happens
    """
    # Create the cache file as a directory to simulate the case where the cache file is
    # not writable
    Path(CACHE_FILE).mkdir(parents=True, exist_ok=False)

    latest_version = check_for_updates()
    assert latest_version is None
    request_get_mock.assert_not_called()


def test_load_last_check_time_ok(fs: FakeFilesystem):
    """
    GIVEN an update cache file
    WHEN load_last_check_time() is called
    THEN it correctly reads the last check time
    """
    Path(CACHE_FILE).parent.mkdir(parents=True, exist_ok=False)
    Path(CACHE_FILE).write_text("check-at: 2")

    assert load_last_check_time() == 2


def test_load_last_check_time_no_cache_ok(caplog, fs: FakeFilesystem):
    """
    GIVEN no update cache file
    WHEN load_last_check_time() is called
    THEN it returns None
    AND no warning messages are logged
    """
    assert load_last_check_time() is None
    assert len(caplog.records) == 0


@pytest.mark.parametrize(
    "content",
    [
        "this-is-not-yaml",
        "no-cache-key: 1",
        "check-at: not_a_float",
    ],
)
def test_load_last_check_time_swallows_errors(caplog, fs: FakeFilesystem, content: str):
    """
    GIVEN invalid content for the cache file
    WHEN load_last_check_time() is called
    THEN it does not raise any exception
    AND it returns None
    AND 1 warning message is logged
    """
    Path(CACHE_FILE).parent.mkdir(parents=True, exist_ok=False)
    Path(CACHE_FILE).write_text(content)

    assert load_last_check_time() is None
    log_record = caplog.records[0]
    assert log_record.levelname == "WARNING"
    assert "Could not load cached latest version" in log_record.message
    assert len(caplog.records) == 1



import json
import urllib.parse as urlparse
from datetime import datetime, timedelta, timezone
from enum import IntEnum, auto
from typing import Optional, Set
from unittest.mock import Mock

import pytest

from ggshield.__main__ import cli
from ggshield.core.config import Config
from ggshield.core.constants import DEFAULT_INSTANCE_URL
from ggshield.core.errors import ExitCode, UnexpectedError
from ggshield.utils.datetime import get_pretty_date
from ggshield.verticals.auth import OAuthClient, OAuthError
from tests.unit.conftest import assert_invoke_ok
from tests.unit.request_mock import (
    RequestMock,
    create_html_response,
    create_json_response,
)

from ..utils import add_instance_config


_EXPECTED_URL_PARAMS = {
    "auth_mode": ["ggshield_login"],
    "client_id": ["ggshield_oauth"],
    "code_challenge_method": ["S256"],
    "response_type": ["code"],
    "utm_campaign": ["ggshield"],
    "utm_medium": ["login"],
    "utm_source": ["cli"],
}

DT_FORMAT = "%Y-%m-%dT%H:%M:%S%z"

TOKEN_ENDPOINT = "/v1/token"

VALID_TOKEN_RESPONSE = create_json_response(
    {
        "type": "personal_access_token",
        "account_id": 17,
        "name": "key",
        "scope": ["scan"],
        "expire_at": None,
    }
)

VALID_TOKEN_INVALID_SCOPE_RESPONSE = create_json_response(
    {
        **VALID_TOKEN_RESPONSE.json(),
        "scope": ["read:incident", "write:incident", "share:incident", "read:member"],
    }
)

INVALID_TOKEN_RESPONSE = create_json_response({"detail": "Invalid API key."}, 401)

METADATA_ENDPOINT = "/v1/metadata"

VALID_METADATA_RESPONSE = create_json_response(
    {
        "version": "1.2.3",
        "preferences": {},
        "secret_scan_preferences": {
            "max_document_size": 1 * 1024 * 1024,
            "max_documents_per_scan": 20,
        },
    }
)


class TestAuthLoginToken:
    @pytest.fixture(autouse=True)
    def setup_method(self, monkeypatch):
        self._request_mock = RequestMock()
        monkeypatch.setattr("ggshield.core.client.Session.request", self._request_mock)

    @pytest.mark.parametrize("test_case", ["valid", "invalid_scope", "invalid"])
    def test_auth_login_token(self, monkeypatch, cli_fs_runner, test_case):
        """
        GIVEN an API token, valid or not
        WHEN the auth login command is called with --method=token
        THEN the validity of the token should be checked, and if valid, the user should be logged in
        """
        token = "mysupertoken"
        instance = "https://dashboard.gitguardian.com"
        cmd = ["auth", "login", "--method=token", f"--instance={instance}"]

        if test_case == "valid":
            self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)
        elif test_case == "invalid_scope":
            self._request_mock.add_GET(
                TOKEN_ENDPOINT, VALID_TOKEN_INVALID_SCOPE_RESPONSE
            )
        elif test_case == "invalid":
            self._request_mock.add_GET(TOKEN_ENDPOINT, INVALID_TOKEN_RESPONSE)

        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        if test_case == "valid":
            assert_invoke_ok(result)
            assert instance in config_instance_urls
            assert config.auth_config.get_instance(instance).account.token == token
        else:
            assert result.exit_code != 0
            if test_case == "invalid_scope":
                assert "This token does not have the scan scope." in result.output
            else:
                assert "Authentication failed with token." in result.output
            assert instance not in config_instance_urls

        self._request_mock.assert_all_requests_happened()

    def test_auth_login_token_default_instance(self, monkeypatch, cli_fs_runner):
        """
        GIVEN a valid API token
        WHEN the auth login command is called without --instance with method token
        THEN the authentication is made against the default instance
        AND ggshield does not suggest calling `ggshield config set instance`
        """
        config = Config()
        assert len(config.auth_config.instances) == 0

        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)

        cmd = ["auth", "login", "--method=token"]

        token = "mysupertoken"
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        assert_invoke_ok(result)
        assert len(config.auth_config.instances) == 1
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        assert config.instance_name in config_instance_urls
        assert (
            config.auth_config.get_instance(config.instance_name).account.token == token
        )
        assert "ggshield config set instance" not in result.output
        self._request_mock.assert_all_requests_happened()

    @pytest.mark.parametrize(
        ("cmd_line_instance", "expected_instance"),
        [
            ("https://api.gitguardian.com/v1", "https://dashboard.gitguardian.com"),
            ("https://api.gitguardian.com", "https://dashboard.gitguardian.com"),
            (
                "https://api.eu1.gitguardian.com/v1",
                "https://dashboard.eu1.gitguardian.com",
            ),
            (
                "https://gitguardian.mycorp.local/exposed/",
                "https://gitguardian.mycorp.local",
            ),
            (
                "https://gitguardian.mycorp.local/exposed",
                "https://gitguardian.mycorp.local",
            ),
        ],
    )
    def test_api_instance_url(
        self, cmd_line_instance, expected_instance, cli_fs_runner
    ):
        """
        GIVEN a valid API token and an instance URL matching GitGuardian API urls
        WHEN running the login command
        THEN it succeeds
        """
        token = "mysupertoken"
        cmd = ["auth", "login", "--method=token", f"--instance={cmd_line_instance}"]
        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")
        config = Config()
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        assert_invoke_ok(result)
        assert expected_instance in config_instance_urls
        self._request_mock.assert_all_requests_happened()

    @pytest.mark.parametrize(
        ("instance", "suggests"),
        (
            ("", False),
            ("https://custom.example.com", True),
            (DEFAULT_INSTANCE_URL, False),
        ),
    )
    def test_auth_login_token_suggests_set_instance(
        self, monkeypatch, cli_fs_runner, instance: str, suggests: bool
    ):
        """
        GIVEN a valid API token
        WHEN the auth login command is called
        THEN ggshield suggests calling `ggshield config set instance` if the instance
             is not the default one
        """
        assert not Config().auth_config.instances

        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)

        cmd = ["auth", "login", "--method=token"]
        if instance:
            cmd.extend(["--instance", instance])

        result = cli_fs_runner.invoke(cli, cmd, color=False, input="a_token\n")
        assert_invoke_ok(result)

        if suggests:
            assert "ggshield config set instance" in result.output
        else:
            assert "ggshield config set instance" not in result.output
        self._request_mock.assert_all_requests_happened()

    def test_auth_login_token_update_existing_config(self, monkeypatch, cli_fs_runner):
        """
        GIVEN some valid API tokens
        WHEN the auth login command is called with --method=token
        THEN the instance configuration is created if it doesn't exist, or updated otherwise
        """
        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)
        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)
        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)

        instance = "https://dashboard.gitguardian.com"
        cmd = ["auth", "login", "--method=token", f"--instance={instance}"]

        token = "myfirstsupertoken"
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        assert_invoke_ok(result)
        assert config.auth_config.get_instance(instance).account.token == token

        token = "mysecondsupertoken"
        result = cli_fs_runner.invoke(cli, cmd, color=False, input=token + "\n")

        config = Config()
        assert_invoke_ok(result)
        assert len(config.auth_config.instances) == 1
        assert config.auth_config.get_instance(instance).account.token == token

        second_instance_token = "mythirdsupertoken"
        second_instance = "https://dashboard.other.gitguardian.com"
        cmd = ["auth", "login", "--method=token", f"--instance={second_instance}"]
        result = cli_fs_runner.invoke(
            cli, cmd, color=False, input=second_instance_token + "\n"
        )

        config = Config()
        assert_invoke_ok(result)
        assert len(config.auth_config.instances) == 2
        assert config.auth_config.get_instance(instance).account.token == token
        assert (
            config.auth_config.get_instance(second_instance).account.token
            == second_instance_token
        )
        self._request_mock.assert_all_requests_happened()

    def test_auth_login_token_from_stdin(self, monkeypatch, cli_fs_runner):
        """
        GIVEN a valid API token in stdin input
        WHEN the auth login command is called
        THEN the authentication is successfully made
        """
        config = Config()
        assert len(config.auth_config.instances) == 0

        self._request_mock.add_GET(TOKEN_ENDPOINT, VALID_TOKEN_RESPONSE)

        token = "mysupertoken"

        get_text_stream_mock = Mock()
        get_text_stream_mock.isatty.return_value = False
        get_text_stream_mock.read.return_value = token
        monkeypatch.setattr(
            "ggshield.cmd.auth.click.get_text_stream",
            Mock(return_value=get_text_stream_mock),
        )

        cmd = ["auth", "login", "--method=token"]

        result = cli_fs_runner.invoke(cli, cmd, color=False)

        config = Config()
        assert_invoke_ok(result)

        assert (
            config.auth_config.get_instance(config.instance_name).account.token == token
        )
        self._request_mock.assert_all_requests_happened()


class LoginResult(IntEnum):
    """Represents the possible results of the `auth login`. Values are ordered in how
    far they can happen in the process.
    """

    NOT_ENOUGH_PORTS = auto()
    INVALID_STATE = auto()
    NO_AUTHORIZATION_CODE = auto()
    EXCHANGE_FAILED = auto()
    GARBAGE_HTML_RESPONSE = auto()
    GARBAGE_NO_TOKEN_RESPONSE = auto()
    INVALID_TOKEN = auto()
    SUCCESS = auto()


class TestAuthLoginWeb:
    @pytest.fixture(autouse=True)
    def setup_method(self, monkeypatch):
        """Define parameter-less mocks. Parametrized mocks are defined in prepare_mocks()."""
        # open browser for the user to login
        self._webbrowser_open_mock = Mock()
        monkeypatch.setattr(
            "ggshield.verticals.auth.oauth.webbrowser.open_new_tab",
            self._webbrowser_open_mock,
        )

        # Ensure that original wait_for_callback method is not called
        self._wait_for_callback_mock = Mock()
        monkeypatch.setattr(
            "ggshield.verticals.auth.oauth.OAuthClient._wait_for_callback",
            self._wait_for_callback_mock,
        )

        self._request_mock = RequestMock()
        monkeypatch.setattr("ggshield.core.client.Session.request", self._request_mock)

        self._token_name = None
        self._lifetime = None
        self._instance_url = None
        self._sso_url = None
        # This is not a list because the --scopes argument takes a single
        # space-separated string
        self._scopes: Optional[str] = None

        config = Config()
        assert len(config.auth_config.instances) == 0

    @pytest.mark.parametrize(
        "instance_url", [DEFAULT_INSTANCE_URL, "https://some_instance.com"]
    )
    def test_existing_token_no_expiry(self, instance_url, cli_fs_runner, monkeypatch):
        self._instance_url = instance_url
        self._request_mock.add_GET(METADATA_ENDPOINT, VALID_METADATA_RESPONSE)

        add_instance_config(instance_url=instance_url)

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output
        self._request_mock.assert_all_requests_happened()
        self._webbrowser_open_mock.assert_not_called()

        self._assert_last_print(
            output, "ggshield is already authenticated without an expiry date\n"
        )

    @pytest.mark.parametrize(
        "instance_url", [DEFAULT_INSTANCE_URL, "https://some_instance.com"]
    )
    @pytest.mark.parametrize(
        ["month", "day", "str_date"],
        [
            ("01", "31", "January 31"),
            ("02", "22", "February 22"),
            ("03", "13", "March 13"),
            ("04", "03", "April 3"),
            ("05", "04", "May 4"),
        ],
    )
    def test_existing_non_expired_token(
        self, month, day, str_date, instance_url, cli_fs_runner, monkeypatch
    ):
        dt = datetime.strptime(f"2100-{month}-{day}T00:00:00+0000", DT_FORMAT)

        self._instance_url = instance_url
        self._request_mock.add_GET(METADATA_ENDPOINT, VALID_METADATA_RESPONSE)

        add_instance_config(instance_url=instance_url, expiry_date=dt)

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output
        self._webbrowser_open_mock.assert_not_called()
        self._request_mock.assert_all_requests_happened()

        self._assert_last_print(
            output, f"ggshield is already authenticated until {str_date}, 2100\n"
        )

    def test_auth_login_recreates_token_if_deleted_server_side(
        self, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN a token stored in the config
        AND the token does not exist on the server
        WHEN `ggshield auth login` is called
        THEN it recreates the token
        """

        # Insert the call to check the stored token
        self._request_mock.add_GET(METADATA_ENDPOINT, INVALID_TOKEN_RESPONSE)
        self.prepare_mocks(monkeypatch)

        add_instance_config()

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output

        self._webbrowser_open_mock.assert_called_once()

        self._request_mock.assert_all_requests_happened()

        assert "Success! You are now authenticated" in output

    def test_no_port_available_exits_error(self, cli_fs_runner, monkeypatch):
        """
        GIVEN -
        WHEN initiating the oauth flow
        AND all potential ports to run the local server are occupied
        THEN the auth flow fails with an explanatory message
        """

        self.prepare_mocks(monkeypatch, login_result=LoginResult.NOT_ENOUGH_PORTS)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.UNEXPECTED_ERROR

        self._webbrowser_open_mock.assert_not_called()
        self._assert_last_print(output, "Error: Could not find unoccupied port.\n")
        self._assert_config_is_empty()

    @pytest.mark.parametrize(
        "login_result",
        [
            LoginResult.INVALID_STATE,
            LoginResult.NO_AUTHORIZATION_CODE,
        ],
    )
    def test_invalid_oauth_params_exits_error(
        self, login_result, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN -
        WHEN receiving the oauth flow callback
        AND the callback doesn't include an authorization code
        OR the state included in the url doesn't match the original state
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(
            monkeypatch,
            login_result=login_result,
        )
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.UNEXPECTED_ERROR

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()
        self._webbrowser_open_mock.assert_called_once()
        self._assert_last_print(
            output,
            "Error: Invalid code or state received from the callback.\n",
        )
        self._assert_config_is_empty()

    def test_invalid_code_exchange_exits_error(self, cli_fs_runner, monkeypatch):
        """
        GIVEN -
        WHEN receiving the oauth flow callback
        AND the call to the server to exchange the authorization code
        against an access token fails
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(monkeypatch, login_result=LoginResult.EXCHANGE_FAILED)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.UNEXPECTED_ERROR

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()

        self._request_mock.assert_all_requests_happened()
        self._webbrowser_open_mock.assert_called_once()
        self._assert_last_print(output, "Error: Cannot create a token: kaboom.\n")

    @pytest.mark.parametrize(
        ("login_result", "message"),
        (
            (
                LoginResult.GARBAGE_HTML_RESPONSE,
                "Error: Server response is not JSON (HTTP code: 418).\n",
            ),
            (
                LoginResult.GARBAGE_NO_TOKEN_RESPONSE,
                "Error: Server did not provide the created token.\n",
            ),
        ),
    )
    def test_garbage_exits_error(
        self, cli_fs_runner, login_result, message, monkeypatch
    ):
        """
        GIVEN a token created via the oauth process
        WHEN the response answer is HTML
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(monkeypatch, login_result=login_result)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.UNEXPECTED_ERROR

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()

        self._request_mock.assert_all_requests_happened()

        self._assert_last_print(output, message)

    def test_invalid_token_exits_error(self, cli_fs_runner, monkeypatch):
        """
        GIVEN a token created via the oauth process
        WHEN the token is invalid
        THEN the auth flow fails with an explanatory message
        """
        self.prepare_mocks(monkeypatch, login_result=LoginResult.INVALID_TOKEN)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.UNEXPECTED_ERROR

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url()

        self._request_mock.assert_all_requests_happened()
        self._assert_last_print(output, "Error: The created token is invalid.\n")

    @pytest.mark.parametrize("token_name", [None, "some token name"])
    @pytest.mark.parametrize("lifetime", [None, 0, 1, 365])
    @pytest.mark.parametrize("used_port_count", [0, 1, 10])
    @pytest.mark.parametrize("existing_expired_token", [False, True])
    @pytest.mark.parametrize("existing_unrelated_token", [False, True])
    @pytest.mark.parametrize("downsized_token", [False, True, None])
    def test_valid_process(
        self,
        downsized_token,
        existing_unrelated_token,
        existing_expired_token,
        used_port_count,
        lifetime,
        token_name,
        cli_fs_runner,
        monkeypatch,
    ):
        self.prepare_mocks(
            monkeypatch,
            token_name=token_name,
            lifetime=lifetime,
            used_port_count=used_port_count,
            downsized_token=downsized_token,
        )

        if existing_expired_token:
            # save expired in config
            add_instance_config(
                expiry_date=datetime.now(tz=timezone.utc).replace(microsecond=0)
                - timedelta(days=2)
            )
        if existing_unrelated_token:
            # add a dummy unrelated config
            add_instance_config(instance_url="http://some-gg-instance.com")

        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url(expected_port=29170 + used_port_count)

        self._request_mock.assert_all_requests_happened()

        if self._lifetime is None:
            str_date = "never"
        else:
            str_date = get_pretty_date(self._get_expiry_date())

        warning_message = ""
        if downsized_token:
            warning_message = (
                " Warning: the expiration date has been adjusted to comply with your workspace's"
                " setting for the maximum lifetime of personal access tokens.\n"
            )

        message = (
            "Success! You are now authenticated.\n"
            "The personal access token has been created and stored in your ggshield config.\n\n"
            f"token name: {self._generated_token_name}\n"
            f"token expiration date: {str_date}\n"
            f"{warning_message}"
            "\n"
            'You do not need to run "ggshield auth login" again. Future requests will automatically use the token.\n'
        )

        assert output.endswith(message)

        self._assert_config("mysupertoken")

    def test_scopes(self, cli_fs_runner, monkeypatch):
        """
        GIVEN a coll to `auth login` with the `--scopes` argument
        WHEN the browser is opened
        THEN the URI includes the scopes
        """
        self.prepare_mocks(monkeypatch, scopes="honeytokens:write teams:read")
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output

        self._webbrowser_open_mock.assert_called_once()
        self._assert_open_url(scope_set={"scan", "honeytokens:write", "teams:read"})

    def prepare_mocks(
        self,
        monkeypatch,
        token_name=None,
        lifetime=None,
        instance_url=None,
        used_port_count=0,
        login_result: LoginResult = LoginResult.SUCCESS,
        sso_url=None,
        downsized_token: Optional[bool] = False,
        scopes: Optional[str] = None,
    ):
        """
        Configure self._request_mock to emulate HTTP requests
        and server interactions.

        Defines the following fields:
        self._token_name
        self._lifetime
        self._instance_url
        self._sso_url
        self._scopes
        self._generated_token_name
        """
        token = "mysupertoken"

        # original token params
        self._token_name = token_name
        self._lifetime = lifetime
        self._instance_url = instance_url
        self._sso_url = sso_url
        self._scopes = scopes

        # token name generated if passed as None
        self._generated_token_name = (
            token_name
            if token_name
            else "ggshield token " + datetime.today().strftime("%Y-%m-%d")
        )

        # generate the expected oauth state
        oauth_state = (
            "invalid_state"
            if login_result == LoginResult.INVALID_STATE
            else self._get_oauth_state()
        )
        url_params = {"state": oauth_state}
        if login_result != LoginResult.NO_AUTHORIZATION_CODE:
            url_params["code"] = "some_authorization_code"

        callback_url = "http://localhost:1234/?" + urlparse.urlencode(url_params)

        # emulates incoming request
        monkeypatch.setattr(
            "ggshield.cmd.auth.login.OAuthClient",
            self._get_oauth_client_class(callback_url),
        )

        # avoid starting a server on port 1234
        if login_result == LoginResult.NOT_ENOUGH_PORTS:
            used_port_count = 1000
        mock_server_class = Mock(
            side_effect=self._get_oserror_side_effect(used_port_count)
        )
        monkeypatch.setattr(
            "ggshield.verticals.auth.oauth.HTTPServer", mock_server_class
        )

        # Step: POST /v1/oauth/token
        if login_result < LoginResult.EXCHANGE_FAILED:
            return

        token_response_payload = {}
        if login_result == LoginResult.EXCHANGE_FAILED:
            response = create_json_response({"detail": "kaboom"}, status_code=400)
        elif login_result == LoginResult.GARBAGE_HTML_RESPONSE:
            response = create_html_response("I'm a teapot", 418)
        elif login_result == LoginResult.GARBAGE_NO_TOKEN_RESPONSE:
            response = create_json_response({"no_key": "nope"})
        elif login_result in (LoginResult.INVALID_TOKEN, LoginResult.SUCCESS):
            token_response_payload = VALID_TOKEN_RESPONSE.json().copy()
            if downsized_token is not None:
                token_response_payload["expire_at_downsized"] = downsized_token
            if lifetime is not None:
                expire_at = self._get_expiry_date().isoformat()
                token_response_payload["expire_at"] = expire_at

            # mock api call to exchange the code against a valid access token
            response = create_json_response({"key": token, **token_response_payload})
        else:
            raise ValueError(f"Invalid {login_result=}")

        self._request_mock.add_POST(
            "/v1/oauth/token", response, self._assert_post_payload
        )

        # GET /v1/token, only if we received a token
        if login_result < LoginResult.INVALID_TOKEN:
            return

        self._request_mock.add_GET(
            TOKEN_ENDPOINT,
            create_json_response(
                token_response_payload,
                400 if login_result == LoginResult.INVALID_TOKEN else 200,
            ),
        )

    def run_cmd(self, cli_fs_runner, method="web"):
        """
        Run the auth login method within a virtual cli.
        Make sure the original server method is not called.
        """
        cmd = ["auth", "login", f"--method={method}"]
        if self._token_name is not None:
            cmd.append(f"--token-name={self._token_name}")

        if self._lifetime is not None:
            cmd.append(f"--lifetime={self._lifetime}")

        if self._instance_url is not None:
            cmd.append(f"--instance={self._instance_url}")

        if self._sso_url is not None:
            cmd.append(f"--sso-url={self._sso_url}")

        if self._scopes is not None:
            cmd.append(f"--scopes={self._scopes}")

        # run cli command
        result = cli_fs_runner.invoke(cli, cmd, color=False, catch_exceptions=False)

        # original method should not be called
        self._wait_for_callback_mock.assert_not_called()
        return result.exit_code, result.output

    @staticmethod
    def _assert_config_is_empty():
        """
        assert that the config is empty
        """
        config = Config()
        assert len(config.auth_config.instances) == 0

    @staticmethod
    def _assert_config(token=None):
        """
        assert that the config exists.
        If a token is passed, assert that the token saved in the config is the same
        """
        config = Config()
        assert len(config.auth_config.instances) >= 1
        config_instance_urls = [
            instance_config.url for instance_config in config.auth_config.instances
        ]
        assert config.instance_name in config_instance_urls
        if token is not None:
            assert (
                config.auth_config.get_instance(config.instance_name).account.token
                == token
            )

    @staticmethod
    def _get_oserror_side_effect(failure_count=1):
        """
        return a side effect to pass to a mock object
        the n first call will raise an exception
        the call n + 1 will be silent
        """
        return (
            OSError("This port is already in use.") if i < failure_count else None
            for i in range(failure_count + 1)
        )

    @staticmethod
    def _assert_last_print(output: str, expected_str: str):
        """
        assert that the last log output is the same as the one passed in param
        """
        assert output.endswith(expected_str)

    def _assert_open_url(
        self,
        *,
        host: Optional[str] = None,
        expected_port: int = 29170,
        scope_set: Optional[Set[str]] = None,
    ):
        """
        assert that the url to be open in the browser has the right static parameters
        also check if the port of the redirect url is the one expected depending on occupied ports
        """
        if scope_set is None:
            scope_set = {"scan"}

        (url,), kwargs = self._webbrowser_open_mock.call_args_list[0]
        parsed_url = urlparse.urlparse(url)
        url_params = urlparse.parse_qs(parsed_url.query)

        if host is not None:
            assert (
                host == parsed_url.netloc
            ), f"invalid host opened: '{parsed_url.netloc}', expected '{host}'"

        for key, value in _EXPECTED_URL_PARAMS.items():
            assert key in url_params, f"missing url param: {key}"
            assert url_params[key] == value, f"invalid value for param '{key}': {value}"

        assert "redirect_uri" in url_params, "redirect_url not in server post request"
        redirect_uri = url_params["redirect_uri"][0]

        assert redirect_uri.startswith(
            f"http://localhost:{expected_port}"
        ), redirect_uri

        # We pass `WebApplicationClient.prepare_request_uri()` a list of
        # strings but it generates a *single* `scope` query parameter, whose
        # value is a space-separated string. That's why we have to split it.
        actual_scope_set = set(url_params["scope"][0].split(" "))
        assert actual_scope_set == scope_set

    def _assert_post_payload(self, payload):
        """
        assert the payload include the expected token name
        """

        request_body = urlparse.parse_qs(payload)

        assert "name" in request_body
        assert request_body["name"][0] == self._generated_token_name

        if self._lifetime is None:
            assert "lifetime" not in request_body
        else:
            assert "lifetime" in request_body
            assert request_body["lifetime"][0] == str(self._lifetime)

    def _get_expiry_date(self) -> Optional[datetime]:
        """
        Get the token expiry date based on the given lifetime
        the date is offset to jan 1st 2100 to make testing easier
        """
        if self._lifetime is not None:
            return datetime.strptime("2100-01-01T00:00:00+0000", DT_FORMAT) + timedelta(
                days=self._lifetime
            )
        return None

    def _get_oauth_state(self):
        return urlparse.quote(
            json.dumps(
                {"token_name": self._generated_token_name, "lifetime": self._lifetime}
            )
        )

    @staticmethod
    def _get_oauth_client_class(callback_url):
        """
        generate a fake OAuth class which overrides the wait for callback function
        instead of waiting for a callback, emulate a callback with the given url
        """

        class FakeOAuthClient(OAuthClient):
            def _wait_for_callback(self, *args, **kwargs):
                try:
                    self.process_callback(callback_url)
                except OAuthError as e:
                    raise UnexpectedError(e.message)

        return FakeOAuthClient

    @pytest.mark.parametrize(
        ["method", "instance_url", "sso_url", "expected_error"],
        [
            [
                "web",
                "https://dashboard.gitguardian.com",
                "https://onprem.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "Error: instance and SSO URL params do not match\n",
            ],
            [
                "web",
                "https://dashboard.gitguardian.com",
                "https://dashboard.gitguardian.com",
                "Error: Invalid value for sso-url: Please provide a valid SSO URL.\n",
            ],
            [
                "token",
                "https://dashboard.gitguardian.com",
                "https://dashboard.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "Error: Invalid value for sso-url: --sso-url is reserved for the web login method.\n",
            ],
        ],
    )
    def test_bad_sso_url(
        self, method, instance_url, sso_url, expected_error, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN an invalid SSO URL, or that do not match the declared instance
        WHEN running the login command
        THEN it fails
        """
        self.prepare_mocks(monkeypatch, instance_url=instance_url, sso_url=sso_url)
        add_instance_config(instance_url=instance_url)

        exit_code, output = self.run_cmd(cli_fs_runner, method=method)
        assert exit_code > 0, output
        self._webbrowser_open_mock.assert_not_called()
        self._assert_last_print(output, expected_error)

    @pytest.mark.parametrize(
        ["instance_url", "sso_url", "expected_web_host"],
        [
            [
                "https://dashboard.gitguardian.com",
                "https://dashboard.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "dashboard.gitguardian.com",
            ],
            [
                "https://some-gg-instance.com",
                "https://some-gg-instance.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "some-gg-instance.com",
            ],
            [
                None,
                "https://custom.gitguardian.com/auth/sso/1e0f7890-2293-4b2d-8aa8-f6f0e8e92274",
                "custom.gitguardian.com",
            ],
            [
                None,
                None,
                "dashboard.gitguardian.com",
            ],
        ],
    )
    def test_sso_url(
        self, instance_url, sso_url, expected_web_host, cli_fs_runner, monkeypatch
    ):
        """
        GIVEN -
        WHEN calling the login command on the web method with a valid SSO URL, or no SSO URL
        THEN if no SSO URL is passed, the default instance is used
        THEN if the SSO URL is compatible with the instance, the instance is used
        THEN if no instance is declared, the instance of the SSO URL is used
        THEN the correct instance URL is checked to see if the auth flow  is enabled
        """
        self.prepare_mocks(monkeypatch, instance_url=instance_url, sso_url=sso_url)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output
        self._webbrowser_open_mock.assert_called()
        self._assert_open_url(host=expected_web_host)

    @pytest.mark.parametrize(
        "instance_url",
        [
            "https://dashboard.gitguardian.com/abc",
        ],
    )
    def test_invalid_instance_url(self, instance_url, cli_fs_runner, monkeypatch):
        """
        GIVEN an invalid instance URL
        WHEN running the login command
        THEN it fails
        """
        monkeypatch.setenv("GITGUARDIAN_INSTANCE", instance_url)

        self.prepare_mocks(monkeypatch)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.USAGE_ERROR, output
        self._webbrowser_open_mock.assert_not_called()

    @pytest.mark.parametrize(
        "instance_url",
        [
            "https://api.gitguardian.com/v1",
            "https://api.eu1.gitguardian.com/v1",
            "https://gitguardian.mycorp.local/exposed/",
        ],
    )
    def test_api_instance_url(self, instance_url, cli_fs_runner, monkeypatch):
        """
        GIVEN an instance URL matching GitGuardian API urls
        WHEN running the login command
        THEN it succeeds
        """
        monkeypatch.setenv("GITGUARDIAN_INSTANCE", instance_url)
        self.prepare_mocks(monkeypatch)
        exit_code, output = self.run_cmd(cli_fs_runner)
        assert exit_code == ExitCode.SUCCESS, output
        self._webbrowser_open_mock.assert_called()



import json
import re
from pathlib import Path
from unittest.mock import Mock, patch

import click
import pytest

from ggshield.__main__ import cli
from ggshield.cmd.secret.scan.secret_scan_common_options import (
    IGNORED_DEFAULT_WILDCARDS,
)
from ggshield.core.errors import ExitCode
from ggshield.core.filter import init_exclusion_regexes
from ggshield.core.scan import StringScannable
from ggshield.verticals.secret import SecretScanCollection
from ggshield.verticals.secret.docker import DockerImage, LayerInfo, _validate_filepath
from tests.unit.conftest import (
    DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH,
    DOCKER_EXAMPLE_PATH,
    UNCHECKED_SECRET_PATCH,
    assert_invoke_exited_with,
    assert_invoke_ok,
    my_vcr,
    write_text,
)


class TestDockerUtils:
    @pytest.mark.parametrize(
        "filepath, exclusion_regexes, valid",
        (
            ["/usr/bin/secret.py", set(), False],
            ["usr/bin/secret.py", set(), False],
            ["/my/file/secret.py", set(), True],
            ["/my/file/usr/bin/secret.py", set(), True],
            ["/usr/share/nginx/secret.py", set(), True],
            ["/gems/secret.py", set(), True],
            ["/npm-bis/secret.py", set(), True],
            ["/banned/extension/secret.exe", set(), False],
            ["/banned/extension/secret.mng", set(), False],
            ["/banned/extension/secret.tar", set(), False],
            ["/banned/extension/secret.other", set(), True],
            [
                "/banned/extension/secret.other",
                {re.compile(r"secret")},
                False,
            ],
        ),
    )
    def test_docker_filepath_validation(self, filepath, exclusion_regexes, valid):
        assert (
            _validate_filepath(
                filepath=filepath,
                exclusion_regexes=exclusion_regexes,
            )
            is valid
        )


class TestDockerCMD:
    @patch("ggshield.cmd.secret.scan.docker.docker_save_to_tmp")
    @patch("ggshield.cmd.secret.scan.docker.docker_scan_archive")
    def test_docker_scan(
        self, scan_mock: Mock, save_mock, cli_fs_runner: click.testing.CliRunner
    ):
        scan_mock.return_value = SecretScanCollection(
            id="ggshield-non-existant", type="docker", results=[]
        )
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "docker", "ggshield-non-existant"],
        )
        assert_invoke_ok(result)

    @patch("ggshield.cmd.secret.scan.docker.docker_save_to_tmp")
    @patch("ggshield.cmd.secret.scan.docker.docker_scan_archive")
    def test_docker_scan_abort(
        self, scan_mock: Mock, save_mock: Mock, cli_fs_runner: click.testing.CliRunner
    ):
        save_mock.side_effect = click.exceptions.Abort()
        scan_mock.return_value = SecretScanCollection(
            id="ggshield-non-existant", type="docker", results=[]
        )
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "docker", "ggshield-non-existant"],
        )
        assert_invoke_ok(result)

        expected_output = ""
        assert result.output == expected_output

    @patch("ggshield.cmd.secret.scan.docker.docker_save_to_tmp")
    @patch("ggshield.cmd.secret.scan.docker.docker_scan_archive")
    def test_docker_scan_failed_to_save(
        self, scan_mock: Mock, save_mock: Mock, cli_fs_runner: click.testing.CliRunner
    ):
        save_mock.side_effect = click.UsageError(
            'Image "ggshield-non-existant" not found'
        )
        scan_mock.return_value = SecretScanCollection(
            id="ggshield-non-existant", type="docker", results=[]
        )
        result = cli_fs_runner.invoke(
            cli,
            ["-v", "secret", "scan", "docker", "ggshield-non-existant"],
        )
        assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)
        assert 'Image "ggshield-non-existant" not found' in result.output

    @patch("ggshield.verticals.secret.docker.DockerImage.open")
    @pytest.mark.parametrize(
        "image_path", [DOCKER_EXAMPLE_PATH, DOCKER__INCOMPLETE_MANIFEST_EXAMPLE_PATH]
    )
    @pytest.mark.parametrize("json_output", (False, True))
    @pytest.mark.parametrize("ignore_secret_file", (True, False))
    def test_docker_scan_archive(
        self,
        docker_image_open_mock: Mock,
        cli_fs_runner: click.testing.CliRunner,
        image_path: Path,
        json_output: bool,
        ignore_secret_file: bool,
    ):
        assert image_path.exists()

        layer_info = LayerInfo(
            filename="12345678/layer.tar", command="COPY foo", diff_id="sha256:1234"
        )

        def create_docker_image() -> Mock(spec=DockerImage):
            docker_image = Mock(spec=DockerImage)
            docker_image.config_scannable = StringScannable(
                content="", url="Dockerfile or build-args"
            )
            docker_image.layer_infos = [layer_info]

            scannable = StringScannable(
                content=UNCHECKED_SECRET_PATCH, url="file_secret"
            )
            docker_image.get_layer_scannables.return_value = (
                [scannable] if not ignore_secret_file else []
            )

            return docker_image

        docker_image = create_docker_image()

        docker_image_open_mock.return_value.__enter__.return_value = docker_image

        with my_vcr.use_cassette(f"test_docker_scan_archive-{ignore_secret_file}"):
            json_arg = ["--json"] if json_output else []
            cli_fs_runner.mix_stderr = False
            if ignore_secret_file:
                config = """
version: 2
secret:
    ignored_paths:
        - "file_secret"

"""
                write_text(Path(".gitguardian.yaml"), config)
            result = cli_fs_runner.invoke(
                cli,
                [
                    "-v",
                    "secret",
                    "scan",
                    *json_arg,
                    "docker-archive",
                    str(image_path),
                ],
            )
            assert_invoke_exited_with(
                result,
                (
                    ExitCode.SCAN_FOUND_PROBLEMS
                    if not ignore_secret_file
                    else ExitCode.SUCCESS
                ),
            )
            docker_image_open_mock.assert_called_once_with(image_path)
            docker_image.get_layer_scannables.assert_called_once_with(
                layer_info,
                init_exclusion_regexes(
                    IGNORED_DEFAULT_WILDCARDS
                    + (["file_secret"] if ignore_secret_file else [])
                ),
            )

            if json_output:
                output = json.loads(result.output)
                if not ignore_secret_file:
                    assert len(output["entities_with_incidents"]) == 1
                else:
                    assert output["total_incidents"] == 0
            else:
                if not ignore_secret_file:
                    assert "file_secret: 1 secret detected" in result.output
                else:
                    assert "No secrets have been found" in result.output



from typing import Any

import click

from ggshield.cmd.hmsl.api_status import status_cmd
from ggshield.cmd.hmsl.check import check_cmd
from ggshield.cmd.hmsl.check_secret_manager import check_secret_manager_group
from ggshield.cmd.hmsl.decrypt import decrypt_cmd
from ggshield.cmd.hmsl.fingerprint import fingerprint_cmd
from ggshield.cmd.hmsl.query import query_cmd
from ggshield.cmd.hmsl.quota import quota_cmd
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.utils.click import NaturalOrderGroup


@click.group(
    cls=NaturalOrderGroup,
    commands={
        "check": check_cmd,
        "check-secret-manager": check_secret_manager_group,
        "fingerprint": fingerprint_cmd,
        "query": query_cmd,
        "decrypt": decrypt_cmd,
        "quota": quota_cmd,
        "api-status": status_cmd,
    },
)
@add_common_options()
def hmsl_group(**kwargs: Any) -> None:
    """Commands for HasMySecretLeaked."""



import hashlib
import operator
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import (
    Counter,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
    cast,
)

from pygitguardian import GGClient
from pygitguardian.models import (
    Detail,
    DiffKind,
    PolicyBreak,
    ScanResult,
    SecretIncident,
)

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.errors import handle_api_error
from ggshield.core.filter import is_in_ignored_matches
from ggshield.core.lines import get_lines_from_content
from ggshield.core.scan import Scannable
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret.extended_match import ExtendedMatch


class IgnoreKind(str, Enum):
    IGNORED_MATCH = "Match ignored via local .gitguardian yaml"
    IGNORED_DETECTOR = "Detector ignored via local .gitguardian yaml"
    KNOWN_SECRET = "Secret is known in dashboard and --ignore-known-secrets is used"
    NOT_INTRODUCED = "Secret was not in added in commit"
    BACKEND_EXCLUDED = "Excluded by dashboard"

    def __str__(self):
        return self.name.lower()


@dataclass(frozen=True)
class IgnoreReason:
    kind: IgnoreKind
    detail: Optional[str] = None

    def to_human_readable(self):
        res = f"{self.kind.value}"
        if self.detail:
            res += f" ({self.detail})"
        return res


def compute_ignore_reason(
    policy_break: PolicyBreak, secret_config: SecretConfig
) -> Optional[IgnoreReason]:
    """Computes the possible ignore reason associated with a PolicyBreak"""
    ignore_reason = None
    if policy_break.diff_kind in {DiffKind.DELETION, DiffKind.CONTEXT}:
        ignore_reason = IgnoreReason(IgnoreKind.NOT_INTRODUCED)
    elif policy_break.is_excluded:
        ignore_reason = IgnoreReason(
            IgnoreKind.BACKEND_EXCLUDED, policy_break.exclude_reason
        )
    elif is_in_ignored_matches(policy_break, secret_config.ignored_matches or []):
        ignore_reason = IgnoreReason(IgnoreKind.IGNORED_MATCH)
    elif policy_break.break_type in secret_config.ignored_detectors:
        ignore_reason = IgnoreReason(IgnoreKind.IGNORED_DETECTOR)
    elif secret_config.ignore_known_secrets and policy_break.known_secret:
        ignore_reason = IgnoreReason(IgnoreKind.KNOWN_SECRET)

    return ignore_reason


@dataclass
class Secret:
    """GGShield specific model to handle policy-breaks.
    Named Secret since we are dropping other kind of policy breaks.
    """

    detector: str
    validity: str
    known_secret: bool
    incident_url: Optional[str]
    matches: List[ExtendedMatch]
    ignore_reason: Optional[IgnoreReason]
    diff_kind: Optional[DiffKind]

    @property
    def policy(self) -> str:
        return "Secrets detection"

    @property
    def is_ignored(self) -> bool:
        return self.ignore_reason is not None

    def get_ignore_sha(self) -> str:
        hashable = "".join(
            [
                f"{match.match},{match.match_type}"
                for match in sorted(self.matches, key=operator.attrgetter("match_type"))
            ]
        )

        return hashlib.sha256(hashable.encode("UTF-8")).hexdigest()


def group_secrets_by_ignore_sha(
    secrets: List[Secret],
) -> Dict[str, List[Secret]]:
    """
    Group policy breaks by their ignore sha.
    """
    sha_dict: Dict[str, List[Secret]] = {}
    for secret in secrets:
        sha_dict.setdefault(secret.get_ignore_sha(), []).append(secret)

    return sha_dict


@dataclass
class Result:
    """
    Return model for a scan which zips the information
    between the Scan result and its input file.
    """

    filename: str  # Name of the file/patch scanned
    filemode: Filemode
    path: Path
    url: str
    secrets: List[Secret]
    ignored_secrets_count_by_kind: Counter[IgnoreKind]

    @property
    def is_on_patch(self) -> bool:
        return self.filemode != Filemode.FILE

    def censor(self) -> None:
        for secret in self.secrets:
            for extended_match in secret.matches:
                cast(ExtendedMatch, extended_match).censor()

    @property
    def has_secrets(self) -> bool:
        return len(self.secrets) > 0

    @classmethod
    def from_scan_result(
        cls, file: Scannable, scan_result: ScanResult, secret_config: SecretConfig
    ) -> "Result":
        """Creates a Result from a Scannable and a ScanResult.
        - Removes ignored policy breaks
        - replace matches by ExtendedMatches
        """

        to_keep: List[Tuple[PolicyBreak, Optional[IgnoreReason]]] = []
        ignored_secrets_count_by_kind = Counter()
        for policy_break in scan_result.policy_breaks:
            ignore_reason = compute_ignore_reason(policy_break, secret_config)
            if ignore_reason is not None:
                if secret_config.all_secrets:
                    to_keep.append((policy_break, ignore_reason))
                else:
                    ignored_secrets_count_by_kind[ignore_reason.kind] += 1
            else:
                to_keep.append((policy_break, None))

        result = Result(
            filename=file.filename,
            filemode=file.filemode,
            path=file.path,
            url=file.url,
            secrets=[],
            ignored_secrets_count_by_kind=ignored_secrets_count_by_kind,
        )

        lines = get_lines_from_content(file.content, file.filemode)
        secrets = [
            Secret(
                validity=policy_break.validity,
                known_secret=policy_break.known_secret,
                incident_url=policy_break.incident_url,
                detector=policy_break.break_type,
                matches=[
                    ExtendedMatch.from_match(match, lines, result.is_on_patch)
                    for match in policy_break.matches
                ],
                ignore_reason=ignore_reason,
                diff_kind=policy_break.diff_kind,
            )
            for policy_break, ignore_reason in to_keep
        ]

        result.secrets = secrets
        return result


class Error(NamedTuple):
    files: List[Tuple[str, Filemode]]
    description: str  # Description of the error


@dataclass
class Results:
    """
    Return model for a scan with the results and errors of the scan

    Not a NamedTuple like the others because it causes mypy 0.961 to crash on the
    `from_exception()` method (!)

    Similar crash: https://github.com/python/mypy/issues/12629
    """

    results: List[Result] = field(default_factory=list)
    errors: List[Error] = field(default_factory=list)

    @staticmethod
    def from_exception(exc: Exception) -> "Results":
        """Create a Results representing a failure"""
        exc_class_name = exc.__class__.__name__
        description = f"{exc_class_name}: {str(exc) or '-'}"
        error = Error(files=[], description=description)
        return Results(results=[], errors=[error])

    def extend(self, others: "Results") -> None:
        self.results.extend(others.results)
        self.errors.extend(others.errors)

    @property
    def has_secrets(self) -> bool:
        return any(x.has_secrets for x in self.results)


class SecretScanCollection:
    id: str
    type: str
    results: Optional[Results] = None
    scans: Optional[List["SecretScanCollection"]] = None
    optional_header: Optional[str] = None  # To be printed in Text Output
    extra_info: Optional[Dict[str, str]] = None  # To be included in JSON Output

    def __init__(
        self,
        id: Union[str, Path],
        type: str,
        results: Optional[Results] = None,
        scans: Optional[List["SecretScanCollection"]] = None,
        optional_header: Optional[str] = None,
        extra_info: Optional[Dict[str, str]] = None,
    ):
        self.id = str(id)
        self.type = type
        self.results = results
        self.scans = scans
        self.optional_header = optional_header
        self.extra_info = extra_info

        self.total_secrets_count = sum(
            len(result.secrets) for result in self.get_all_results()
        )

    @property
    def scans_with_results(self) -> List["SecretScanCollection"]:
        if self.scans:
            return [scan for scan in self.scans if scan.results]
        return []

    def get_all_results(self) -> Iterable[Result]:
        """Returns an iterable on all results and sub-scan results"""
        if self.results:
            yield from self.results.results
        if self.scans:
            for scan in self.scans:
                if scan.results:
                    yield from scan.results.results

    def get_incident_details(self, client: GGClient) -> Dict[str, SecretIncident]:
        incident_details: dict[str, SecretIncident] = {}
        for result in self.get_all_results():
            for secret in result.secrets:
                url = secret.incident_url
                if url and url not in incident_details:
                    incident_id = int(url.split("/")[-1])
                    resp = client.retrieve_secret_incident(
                        incident_id, with_occurrences=0
                    )
                    if isinstance(resp, SecretIncident):
                        incident_details[url] = resp
                    else:
                        assert isinstance(resp, Detail)
                        handle_api_error(resp)
        return incident_details



import re
import tempfile
from pathlib import Path
from typing import Any

import click
from click import UsageError

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.git_shell import git
from ggshield.verticals.secret.repo import scan_repo_path


# Source: https://github.com/jonschlinkert/is-git-url MIT LICENSE
# TODO: it should be possible to scan a repo URL which does not end with ".git"
REGEX_GIT_URL = re.compile(
    r"(?:git|ssh|https?|git@[-\w.]+):(//)?(.*?)(\.git)(/?|#[-\d\w._]+?)$"
)


@click.command()
@click.argument("repository", nargs=1, type=click.STRING, required=True)
@add_secret_scan_common_options()
@click.pass_context
def repo_cmd(
    ctx: click.Context, repository: str, **kwargs: Any
) -> int:  # pragma: no cover
    """
    Scan a REPOSITORY's commits at the given URL or path.

    REPOSITORY is the clone URL or the path of the repository to scan.
    """
    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    cache = ctx_obj.cache
    ctx_obj.client = create_client_from_config(config)

    scan_context = ScanContext(
        scan_mode=ScanMode.REPO,
        command_path=ctx.command_path,
    )

    path = Path(repository)
    if path.is_dir():
        scan_context.target_path = path
        return scan_repo_path(
            client=ctx_obj.client,
            cache=cache,
            output_handler=create_output_handler(ctx),
            exclusion_regexes=ctx_obj.exclusion_regexes,
            config=config,
            scan_context=scan_context,
            repo_path=path,
        )

    if REGEX_GIT_URL.match(repository):
        with tempfile.TemporaryDirectory() as tmpdirname:
            git(["clone", "--mirror", repository, tmpdirname])
            scan_context.target_path = Path(tmpdirname)
            return scan_repo_path(
                client=ctx_obj.client,
                cache=cache,
                output_handler=create_output_handler(ctx),
                exclusion_regexes=ctx_obj.exclusion_regexes,
                config=config,
                scan_context=scan_context,
                repo_path=Path(tmpdirname),
            )

    if any(host in repository for host in ("gitlab.com", "github.com")):
        raise UsageError(
            f"{repository} doesn't seem to be a valid git URL.\n"
            f"Did you mean {repository}.git?"
        )
    raise UsageError(f"{repository} is neither a valid path nor a git URL")



import logging
from typing import Optional

from . import log


FILE_LOG_FORMAT = "%(asctime)s %(levelname)s %(process)x:%(thread)x %(name)s:%(funcName)s:%(lineno)d %(message)s"


_log_handler: Optional[logging.Handler] = None


class _LogHandler(logging.Handler):
    """Implements logging.Handler to pass the log record to ui.log()"""

    def emit(self, record: logging.LogRecord) -> None:
        log(record)


def disable_logs() -> None:
    """By default, disable all logs, because when an error occurs we log an error
    message and also print a human-friendly message using display_errors(). If we don't
    disable all logs, then error logs are printed, resulting in the error being shown
    twice.
    """
    logging.disable()


def set_log_handler(filename: Optional[str] = None) -> None:
    """Configure Python logger to log to stderr if filename is None, or to filename if
    it's set.
    """
    global _log_handler

    # Re-enable logging, reverting any call to disable_logs()
    logging.disable(logging.NOTSET)

    if filename and filename != "-":
        _log_handler = logging.FileHandler(filename)
        _log_handler.setFormatter(logging.Formatter(FILE_LOG_FORMAT))
    else:
        _log_handler = _LogHandler()

    logging.basicConfig(level=logging.DEBUG, force=True, handlers=[_log_handler])


def _reset_log_handler():
    """Remove our log handler. Used by reset.reset()."""
    global _log_handler
    if _log_handler:
        logging.getLogger().removeHandler(_log_handler)
        _log_handler = None



from pathlib import Path
from typing import Pattern, Set

from click import UsageError

from ggshield.utils.files import is_path_excluded


def check_directory_not_ignored(directory: Path, exclusion_regexes: Set[Pattern[str]]):
    if is_path_excluded(directory.resolve(), exclusion_regexes):
        raise UsageError("An ignored file or directory cannot be scanned.")



import logging
from typing import Any, TextIO, cast

import click

from ggshield.cmd.hmsl.hmsl_common_options import (
    full_hashes_option,
    input_arg,
    input_type_option,
    naming_strategy_option,
)
from ggshield.cmd.hmsl.hmsl_utils import check_secrets
from ggshield.cmd.utils.common_options import (
    add_common_options,
    json_option,
    text_json_format_option,
)
from ggshield.core import ui
from ggshield.verticals.hmsl.collection import (
    InputType,
    NamingStrategy,
    collect,
    prepare,
)


logger = logging.getLogger(__name__)


@click.command()
@click.pass_context
@add_common_options()
@text_json_format_option
@json_option
@full_hashes_option
@naming_strategy_option
@input_type_option
@input_arg
def check_cmd(
    ctx: click.Context,
    path: str,
    full_hashes: bool,
    naming_strategy: NamingStrategy,
    input_type: InputType,
    **kwargs: Any,
) -> int:
    """
    Check if secrets have leaked.

    Note: Secrets can be read from stdin using `ggshield hmsl check -`.
    """

    # Collect secrets
    ui.display_info("Collecting secrets...")
    input = cast(TextIO, click.open_file(path, "r"))
    secrets = list(collect(input, input_type))
    # full_hashes is True because we need the hashes to decrypt the secrets.
    # They will correctly be truncated by our client later.
    prepared_data = prepare(secrets, naming_strategy, full_hashes=True)
    ui.display_info(f"Collected {len(prepared_data.payload)} secrets.")

    check_secrets(
        ctx=ctx,
        prepared_secrets=prepared_data,
        full_hashes=full_hashes,
    )

    return 0



import random
import string


def generate_random_honeytoken_name() -> str:
    """Generate a honeytoken name based on a random string of eight alphanumeric characters"""
    letters_and_digits = string.ascii_letters + string.digits
    random_str = "".join(random.choice(letters_and_digits) for i in range(8))
    return f"ggshield-{random_str}"



from collections import Counter
from pathlib import Path
from typing import Any, List, Optional

import click
from pygitguardian.models import Detail, HoneytokenWithContextResponse

from ggshield.cmd.honeytoken.utils import generate_random_honeytoken_name
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.errors import UnexpectedError
from ggshield.utils.click import RealPath


def _list_extensions_in_dir(dir: Path, max_depth: int = 3, depth: int = 0) -> List[str]:
    ext_list = []
    for path in dir.glob("*"):
        if path.is_file() and path.suffix:
            ext_list.append(path.suffix)
        if path.is_dir() and depth < max_depth:
            ext_list.extend(_list_extensions_in_dir(path, max_depth, depth + 1))
    return ext_list


def _get_most_common_extensions_in_dir(dir: Path):
    """
    Return the most 3 common extensions in a directory
    """
    counter = Counter(_list_extensions_in_dir(dir))
    return [extension for extension, _ in counter.most_common(3)]


@click.command()
@click.option(
    "--name",
    "name",
    required=False,
    type=str,
    help="Specify a name for your honeytoken. If this option is not provided, a unique name will be generated with a \
`ggshield-` prefix.",
)
@click.option(
    "--type",
    "type_",
    required=True,
    type=click.Choice(("AWS",)),
    help="Specify the type of honeytoken that you want to create. (For now only AWS honeytokens are supported)",
)
@click.option(
    "--description",
    "description",
    required=False,
    type=str,
    help="Add a description to your honeytoken (250 characters max).",
)
@click.option(
    "--language",
    "language",
    required=False,
    type=str,
    help="Language to use for the context. If not set, ggshield infers the language \
        from the repository or from OUTPUT_FILE, if set.",
)
@click.option(
    "-o",
    "--output",
    "output_file",
    type=RealPath(file_okay=True, dir_okay=False, readable=True, writable=True),
    required=False,
    help="Filename to store your honeytoken.",
)
@add_common_options()
@click.pass_context
def create_with_context_cmd(
    ctx: click.Context,
    name: Optional[str],
    type_: str,
    description: Optional[str],
    output_file: Optional[Path],
    language: Optional[str],
    **kwargs: Any,
) -> int:
    """
    Create a honeytoken within a context.

    The context is a realistic file in which your honeytoken is inserted.
    Adding your honeytoken within a relevant context makes it look more credible.

    The prerequisites to use this command are the following:

    - you have the necessary permissions as a user (for now, Honeytoken is
    restricted to users with a "Manager" access level),

    - the personal access token used by ggshield has the `honeytokens:write`
    scope.
    """
    # if name is not given, generate a random one
    if not name:
        name = generate_random_honeytoken_name()
    ctx_obj = ContextObj.get(ctx)
    client = create_client_from_config(ctx_obj.config)

    response = client.create_honeytoken_with_context(
        name=name,
        type_=type_,
        description=description,
        language=language,
        filename=output_file.name if output_file else None,
        project_extensions=(
            []
            if language or output_file
            else _get_most_common_extensions_in_dir(Path("."))
        ),
    )

    if not isinstance(response, (Detail, HoneytokenWithContextResponse)):
        raise UnexpectedError("Unexpected honeytoken response")

    if (
        isinstance(response, Detail)
        and response.status_code == 403
        and "allowlist" not in response.detail
    ):
        raise UnexpectedError(
            """ggshield does not have permissions to create honeytokens on your workspace. Make sure that:

- the honeytoken module is enabled for your GitGuardian workspace,
- you have the necessary permissions as a user,
- the personal access token used by ggshield has the `honeytokens:write` scope.

To learn more, visit https://docs.gitguardian.com/honeytoken/getting-started."""
        )
    elif isinstance(response, Detail):
        raise UnexpectedError(response.detail)

    filepath_to_write = Path(response.filename)

    click.echo(
        "Honeytoken created and placed in the following context:\n"
        f"- file: {filepath_to_write.absolute()}\n"
        f"- language: {response.language}\n"
        "\n"
        f'Your honeytoken "{name}" is accessible in your workspace: {response.gitguardian_url}'
    )

    filepath_to_write.write_text(response.content)

    return 0



import abc
import http.server
import shutil
import socketserver
import time
from multiprocessing import Process
from pathlib import Path
from typing import Generator
from urllib.parse import urlparse

import pytest
import requests
from pygitguardian.config import DEFAULT_BASE_URI

from tests.repository import Repository


GGSHIELD_PATH = shutil.which("ggshield")

FUNCTESTS_DATA_PATH = Path(__file__).parent / "data"

# Path to the root of ggshield repository
REPO_PATH = Path(__file__).parent.parent.parent

HAS_DOCKER = shutil.which("docker") is not None

HOOK_CONTENT = """#!/usr/bin/env sh
ggshield {} scan pre-receive
"""

HOOK_CONTENT_ALL = """#!/usr/bin/env sh
ggshield {} scan pre-receive --all
"""


# Use this as a decorator for tests which call the `docker` binary
requires_docker = pytest.mark.skipif(not HAS_DOCKER, reason="This test requires Docker")


class AbstractGGAPIHandler(http.server.BaseHTTPRequestHandler, metaclass=abc.ABCMeta):
    def do_HEAD(self):
        self.send_response(200)

    def do_GET(self):
        # Forward all GET calls to the real server
        url = DEFAULT_BASE_URI + self.path.replace("/exposed", "")
        headers = {
            **self.headers,
            "Host": urlparse(url).netloc,
        }

        response = requests.get(url, headers=headers)

        self.send_response(response.status_code)

        for name, value in response.headers.items():
            if name != "content-encoding" and name != "transfer-encoding":
                # Forward headers, but not content-encoding nor transfer-encoding
                # because our response is not compressed and/or chunked content, even if
                # we received it that way
                self.send_header(name, value)
        self.end_headers()

        self.wfile.write(response.content)

    @abc.abstractmethod
    def do_POST(self):
        raise NotImplementedError()


class SlowGGAPIHandler(AbstractGGAPIHandler):
    def do_POST(self):
        if "multiscan" in self.path:
            content = b'{"detail":"Sorry, I overslept!"}'
            self.send_response(200)
            self.send_header("content-type", "application/json")
            self.send_header("Content-Length", str(len(content)))
            self.end_headers()
            time.sleep(60)
            self.wfile.write(content)
        else:
            self.send_response(418)


class NoQuotaGGAPIHandler(AbstractGGAPIHandler):
    def do_POST(self):
        content = b'{"detail":"Quota limit reached."}'
        self.send_response(403)
        self.send_header("content-type", "application/json")
        self.send_header("Content-Length", str(len(content)))
        self.end_headers()
        self.wfile.write(content)


class ReuseAddressServer(socketserver.TCPServer):
    allow_reuse_address = True


def _start_slow_gitguardian_api(host: str, port: int):
    with ReuseAddressServer((host, port), SlowGGAPIHandler) as httpd:
        httpd.serve_forever()


def _start_no_quota_gitguardian_api(host: str, port: int):
    with ReuseAddressServer((host, port), NoQuotaGGAPIHandler) as httpd:
        httpd.serve_forever()


@pytest.fixture
@pytest.mark.allow_hosts(["localhost"])
def slow_gitguardian_api() -> Generator[str, None, None]:
    host, port = "localhost", 8123
    server_process = Process(target=_start_slow_gitguardian_api, args=(host, port))
    server_process.start()
    try:
        yield f"http://{host}:{port}"
    finally:
        server_process.kill()
        server_process.join()


@pytest.fixture
@pytest.mark.allow_hosts(["localhost"])
def no_quota_gitguardian_api() -> Generator[str, None, None]:
    host, port = "localhost", 8124
    server_process = Process(target=_start_no_quota_gitguardian_api, args=(host, port))
    server_process.start()
    try:
        yield f"http://{host}:{port}"
    finally:
        server_process.kill()
        server_process.join()


def repo_with_hook_content(tmp_path: Path, hook_content: str) -> Repository:
    """
    Helper function that initialize a repo with a remote.
    The remote contains the pre-receive with the corresponding hook content.

    :param tmp_path: the root path
    :param hook_content: the pre-receive hook content
    :return: the local Repository object
    """
    remote_repo = Repository.create(tmp_path / "remote", bare=True)
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")

    hook_path = remote_repo.path / "hooks" / "pre-receive"
    hook_path.write_text(hook_content)
    hook_path.chmod(0o700)
    return local_repo


def pytest_report_header(config, start_path: Path):
    """This function is called by pytest, it lets us insert messages in its report
    header"""
    return f"ggshield path: {GGSHIELD_PATH}"



class VaultCliTokenFetchingError(Exception):
    """Raised when the token used by Vault CLI cannot be fetched."""

    pass


class VaultInvalidUrlError(Exception):
    """Raised when the Vault instance URL cannot be parsed."""

    pass


class VaultPathIsNotADirectoryError(Exception):
    """Raised when list_kv_items was called on a file and not a directory."""

    pass


class VaultNotFoundItemError(Exception):
    """Raised when list_kv_items was called on a directory and not a file."""

    pass


class VaultForbiddenItemError(Exception):
    """Raised when a 403 forbidden error was returned for an item."""

    pass



import os
from pathlib import Path
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner

from ggshield.__main__ import cli
from ggshield.cmd.install import get_default_global_hook_dir_path
from ggshield.core.errors import ExitCode
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok


SAMPLE_PRE_COMMIT = """#!/bin/sh
ggshield secret scan pre-commit "$@"
"""

SAMPLE_PRE_PUSH = """#!/bin/sh
ggshield secret scan pre-push "$@"
"""


class TestInstallLocal:
    def test_local_exist_is_dir(self, cli_fs_runner):
        os.system("git init")
        hook_path = Path(".git/hooks/pre-commit")
        hook_path.mkdir()

        result = cli_fs_runner.invoke(cli, ["install", "-m", "local"])
        assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)
        assert result.exception
        assert f"Error: {hook_path} is a directory" in result.output

    def test_local_exist_not_force(self, cli_fs_runner):
        os.system("git init")
        hook_path = Path(".git/hooks/pre-commit")
        hook_path.write_text("pre-commit file")

        result = cli_fs_runner.invoke(cli, ["install", "-m", "local"])
        assert_invoke_exited_with(result, ExitCode.UNEXPECTED_ERROR)
        assert result.exception
        assert f"Error: {hook_path} already exists." in result.output

    def test_local_exist_force(self, cli_fs_runner):
        os.system("git init")
        hook_path = Path(".git/hooks/pre-commit")
        hook_path.write_text("pre-commit file")

        result = cli_fs_runner.invoke(cli, ["install", "-f", "-m", "local"])
        assert_invoke_ok(result)
        assert f"pre-commit successfully added in {hook_path}" in result.output

    @patch("ggshield.cmd.install.check_git_dir")
    def test_precommit_install(
        self,
        check_dir_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN None
        WHEN the command is run
        THEN it should create a pre-commit git hook script
        """

        result = cli_fs_runner.invoke(
            cli,
            ["install", "-m", "local"],
        )
        hook_path = Path(".git/hooks/pre-commit")
        hook_str = hook_path.read_text()
        assert hook_str == SAMPLE_PRE_COMMIT

        assert f"pre-commit successfully added in {hook_path}\n" in result.output
        assert_invoke_ok(result)

    @pytest.mark.parametrize("hook_type", ["pre-push", "pre-commit"])
    @patch("ggshield.cmd.install.check_git_dir")
    def test_install_exists(
        self,
        check_dir_mock: Mock,
        hook_type: str,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a hook that already exists
        WHEN the command is run without --force or --append
        THEN it should error
        """
        hook_path = Path(".git/hooks") / hook_type
        hook_path.parent.mkdir(parents=True)
        hook_path.write_text("#!/bin/bash\nsample-command\n")

        result = cli_fs_runner.invoke(
            cli,
            ["install", "-m", "local", "-t", hook_type],
        )

        assert (
            "already exists. Use --force to override or --append to add to current script\n"
            in result.output
        )
        assert_invoke_exited_with(result, ExitCode.UNEXPECTED_ERROR)

    @pytest.mark.parametrize("hook_type", ["pre-push", "pre-commit"])
    @patch("ggshield.cmd.install.check_git_dir")
    def test_install_exists_force(
        self,
        check_dir_mock: Mock,
        hook_type: str,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a hook that already exists
        WHEN the command is run with --force
        THEN it should return 0 and install the hook
        """
        hook_path = Path(".git/hooks") / hook_type
        hook_path.parent.mkdir(parents=True)
        hook_path.write_text("#!/bin/bash\nsample-command\n")

        result = cli_fs_runner.invoke(
            cli,
            ["install", "-m", "local", "-t", hook_type, "--force"],
        )

        assert f"{hook_type} successfully added in {hook_path}\n" in result.output
        assert_invoke_ok(result)

    @pytest.mark.parametrize("hook_type", ["pre-push", "pre-commit"])
    @patch("ggshield.cmd.install.check_git_dir")
    def test_install_exists_append(
        self,
        check_dir_mock: Mock,
        hook_type: str,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN a hook that already exists
        WHEN the command is run with --append
        THEN it should return 0 and append the hook to the existing one
        """
        hook_path = Path(".git/hooks") / hook_type
        hook_path.parent.mkdir(parents=True)
        hook_path.write_text("#!/bin/bash\nsample-command\n")

        result = cli_fs_runner.invoke(
            cli,
            ["install", "-m", "local", "-t", hook_type, "--append"],
        )
        hook_str = hook_path.read_text()
        assert "sample-command" in hook_str
        assert "ggshield secret scan" in hook_str

        assert f"{hook_type} successfully added in {hook_path}\n" in result.output
        assert_invoke_ok(result)

    @patch("ggshield.cmd.install.check_git_dir")
    def test_prepush_install(
        self,
        check_dir_mock: Mock,
        cli_fs_runner: CliRunner,
    ):
        """
        GIVEN None
        WHEN the command is run
        THEN it should create a pre-push git hook script
        """

        result = cli_fs_runner.invoke(
            cli,
            ["install", "-m", "local", "-t", "pre-push"],
        )
        hook_path = Path(".git/hooks/pre-push")
        hook_str = hook_path.read_text()
        assert hook_str == SAMPLE_PRE_PUSH

        assert f"pre-push successfully added in {hook_path}\n" in result.output
        assert_invoke_ok(result)


@pytest.fixture()
def custom_global_git_config_path(tmp_path, monkeypatch):
    config_path = tmp_path / "global-git-config"
    monkeypatch.setenv("GIT_CONFIG_GLOBAL", str(config_path))
    yield config_path


class TestInstallGlobal:
    """
    These tests use the cli_runner fixture and not the cli_fs_runner one. The reason for
    this is they execute git commands and git commands are not run in the fake
    filesystem created by cli_fs_runner so the fake filesystem is useless here.
    """

    def test_global_exist_is_dir(
        self, cli_runner: CliRunner, custom_global_git_config_path: Path
    ):
        global_pre_commit_hook_path = get_default_global_hook_dir_path() / "pre-commit"
        global_pre_commit_hook_path.mkdir(parents=True)

        result = cli_runner.invoke(cli, ["install", "-m", "global"])
        assert_invoke_exited_with(result, ExitCode.USAGE_ERROR)
        assert result.exception

    def test_global_not_exist(self, cli_runner, custom_global_git_config_path: Path):
        global_pre_commit_hook_path = get_default_global_hook_dir_path() / "pre-commit"
        assert not global_pre_commit_hook_path.exists()

        result = cli_runner.invoke(cli, ["install", "-m", "global"])
        assert global_pre_commit_hook_path.is_file()
        assert_invoke_ok(result)
        assert (
            f"pre-commit successfully added in {global_pre_commit_hook_path}"
            in result.output
        )

    def test_install_custom_global_hook_dir(
        self, cli_runner: CliRunner, tmp_path: Path, custom_global_git_config_path: Path
    ):
        """
        GIVEN an existing global git config
        AND a custom value for core.hooksPath in the global git config
        WHEN `install -m global` is called
        THEN it installs the hook in the custom core.hooksPath dir
        """
        custom_hooks_dir = tmp_path / "custom-hooks-dir"
        custom_pre_commit_path = custom_hooks_dir / "pre-commit"
        custom_global_git_config_path.write_text(
            f"[core]\nhooksPath = {custom_hooks_dir.as_posix()}\n", encoding="utf-8"
        )

        result = cli_runner.invoke(cli, ["install", "-m", "global"])
        assert_invoke_ok(result)
        assert custom_pre_commit_path.is_file()
        assert (
            f"pre-commit successfully added in {custom_pre_commit_path}"
            in result.output
        )

    @pytest.mark.parametrize("hook_type", ["pre-push", "pre-commit"])
    def test_install_global(
        self,
        hook_type: str,
        cli_runner: CliRunner,
        custom_global_git_config_path: Path,
    ):
        """
        GIVEN None
        WHEN the command is run
        THEN it should create a pre-push git hook script in the global path
        """

        result = cli_runner.invoke(
            cli,
            ["install", "-m", "global", "-t", hook_type],
        )

        hook_path = get_default_global_hook_dir_path() / hook_type
        hook_str = hook_path.read_text()
        assert f"if [ -f .git/hooks/{hook_type} ]; then" in hook_str
        assert f"ggshield secret scan {hook_type}" in hook_str

        assert f"{hook_type} successfully added in {hook_path}\n" in result.output
        assert_invoke_ok(result)

    def test_global_exist_not_force(
        self, cli_runner: CliRunner, custom_global_git_config_path: Path
    ):
        """
        GIVEN a global hook dir with an exising pre-commit script
        WHEN install is called
        THEN it fails
        """
        hook_path = get_default_global_hook_dir_path() / "pre-commit"
        hook_path.parent.mkdir(parents=True, exist_ok=True)
        hook_path.write_text("pre-commit file")
        assert hook_path.is_file()

        result = cli_runner.invoke(cli, ["install", "-m", "global"])
        assert_invoke_exited_with(result, ExitCode.UNEXPECTED_ERROR)
        assert result.exception
        assert f"Error: {hook_path} already exists." in result.output

    def test_global_exist_force(
        self, cli_runner: CliRunner, custom_global_git_config_path: Path
    ):
        """
        GIVEN a global hook dir with an exising pre-commit script
        WHEN install is called with -f
        THEN it ignores the fact that the pre-commit script exists and succeeds
        """
        hook_path = get_default_global_hook_dir_path() / "pre-commit"
        hook_path.parent.mkdir(parents=True, exist_ok=True)
        hook_path.write_text("pre-commit file")
        assert hook_path.is_file()

        result = cli_runner.invoke(cli, ["install", "-m", "global", "-f"])
        assert_invoke_ok(result)
        assert f"pre-commit successfully added in {hook_path}" in result.output



import json
from typing import Any, Iterable, Iterator, List, TextIO

import click

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.client import create_client_from_config
from ggshield.core.scan import ScanContext, ScanMode, Scannable, StringScannable
from ggshield.core.scanner_ui import create_message_only_scanner_ui
from ggshield.core.ui.ggshield_ui import GGShieldProgress
from ggshield.verticals.secret import SecretScanCollection, SecretScanner


def generate_files_from_docsets(file: TextIO) -> Iterator[Scannable]:
    for line in file:
        obj = json.loads(line)
        documents = obj["documents"]
        for document in documents:
            ui.display_verbose(f"  * {document['id']}")
            yield StringScannable(document["id"], document["content"])


def create_scans_from_docset_files(
    scanner: SecretScanner, input_files: Iterable[TextIO], progress: GGShieldProgress
) -> List[SecretScanCollection]:
    scans: List[SecretScanCollection] = []

    for input_file in input_files:
        ui.display_verbose(f"- {click.format_filename(input_file.name)}")

        files = generate_files_from_docsets(input_file)
        with create_message_only_scanner_ui() as scanner_ui:
            results = scanner.scan(files, scanner_ui=scanner_ui)
        scans.append(
            SecretScanCollection(id=input_file.name, type="docset", results=results)
        )
        progress.advance(1)

    return scans


@click.command()
@click.argument("files", nargs=-1, type=click.File(), required=True)
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def docset_cmd(
    ctx: click.Context,
    files: List[TextIO],
    **kwargs: Any,
) -> int:  # pragma: no cover
    """
    Scan docset JSONL files.

    The JSONL files must be formatted using the ["Docset" format][1].

    \b
    [1]: https://docs.gitguardian.com/ggshield-docs/integrations/other-data-sources/other-data-sources
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config
    output_handler = create_output_handler(ctx)
    with ui.create_progress(len(files)) as progress:

        scan_context = ScanContext(
            scan_mode=ScanMode.DOCSET,
            command_path=ctx.command_path,
        )
        scanner = SecretScanner(
            client=ctx_obj.client,
            cache=ctx_obj.cache,
            secret_config=config.user_config.secret,
            scan_context=scan_context,
        )
        scans = create_scans_from_docset_files(
            scanner=scanner, input_files=files, progress=progress
        )

    return output_handler.process_scan(
        SecretScanCollection(id=scan_context.command_id, type="docset", scans=scans)
    )



from ggshield.core.filter import censor_match
from ggshield.core.text_utils import pluralize, translate_validity

from ..secret_scan_collection import Secret, SecretScanCollection
from .secret_output_handler import SecretOutputHandler


def format_secret(secret: Secret) -> str:
    """Returns a string with the policy name, validity and a comma-separated,
    double-quoted, censored version of all `secret` matches.

    Looks like this:

    PayPal OAuth2 Keys (Validity: Valid, id="aa*******bb", secret="cc******dd")
    """
    match_str = ", ".join(
        f'{x.match_type}: "{censor_match(x)}"' for x in secret.matches
    )
    validity = translate_validity(secret.validity)
    return f"{secret.detector} (Validity: {validity}, {match_str})"


class SecretGitLabWebUIOutputHandler(SecretOutputHandler):
    """
    Terse OutputHandler optimized for GitLab Web UI, because GitLab Web UI only shows
    lines starting with GL-HOOK-ERR.

    See https://docs.gitlab.com/ee/administration/server_hooks.html#custom-error-messages
    """

    use_stderr = True

    def _process_scan_impl(self, scan: SecretScanCollection) -> str:
        results = scan.get_all_results()

        secrets_to_report = [secret for result in results for secret in result.secrets]

        # If no secrets or no new secrets were found
        if len(secrets_to_report) == 0:
            return ""

        # Use a set to ensure we do not report duplicate incidents.
        # (can happen when the secret is present in both the old and the new version of
        # the document)
        formatted_secrets = {format_secret(x) for x in secrets_to_report}
        break_count = len(formatted_secrets)

        if self.ignore_known_secrets:
            summary_str = f"{break_count} new {pluralize('incident', break_count)}"
        else:
            summary_str = f"{break_count} {pluralize('incident', break_count)}"

        # Putting each policy break on its own line would be more readable, but we can't
        # do this because of a bug in GitLab Web IDE which causes newline characters to
        # be shown as "<br>"
        # https://gitlab.com/gitlab-org/gitlab/-/issues/350349
        breaks_str = ", ".join(formatted_secrets)

        return (
            f"GL-HOOK-ERR: ggshield found {summary_str} in these changes: {breaks_str}."
            " The commit has been rejected."
        )



from typing import Any, TextIO, cast

import click

from ggshield.cmd.hmsl.hmsl_common_options import (
    full_hashes_option,
    input_arg,
    input_type_option,
    naming_strategy_option,
)
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.core import ui
from ggshield.verticals.hmsl.collection import (
    InputType,
    NamingStrategy,
    collect,
    prepare,
)
from ggshield.verticals.hmsl.output import write_outputs


def validate_prefix(prefix: str) -> str:
    if prefix != "" and not prefix.endswith("-"):
        prefix = prefix + "-"
    return prefix


@click.command()
@add_common_options()
@click.option(
    "--prefix",
    "-p",
    default="",
    help="Prefix for output file names. For instance `-p foo` produces `foo-payload.txt` and `foo-mapping.txt`.",
    callback=lambda _, __, value: validate_prefix(value),
    metavar="PREFIX",
)
@full_hashes_option
@naming_strategy_option
@input_type_option
@input_arg
def fingerprint_cmd(
    path: str,
    prefix: str,
    naming_strategy: NamingStrategy,
    full_hashes: bool,
    input_type: InputType,
    **_: Any,
) -> int:
    """
    Collect secrets and compute fingerprints.

    Fingerprints are to be used later by the `decrypt` command.

    Note: Secrets can be read from stdin using `ggshield hmsl fingerprint -`.
    """
    # Opens the file or stdin
    input = cast(TextIO, click.open_file(path, "r"))

    # Prepare and write the output files
    secrets = list(collect(input, input_type))
    result = prepare(secrets, naming_strategy, full_hashes=full_hashes)
    write_outputs(result, prefix)

    ui.display_info(f"Prepared {len(secrets)} secrets.")
    return 0



from pathlib import Path

import pytest

from ggshield.core.scan import StringScannable


def test_string_scannable_path():
    """
    GIVEN a StringScannable instance
    WHEN path() is called
    THEN it returns the right value
    """
    scannable = StringScannable(url="custom:/some/path", content="")
    assert scannable.path == Path("/some/path")


@pytest.mark.parametrize(
    ("content", "is_longer"),
    (
        ("x" * 100, True),
        ("x" * 10, False),
        ("é" * 40, True),  # Longer than 50 as utf-8
        ("\uD800", False),  # Triggers an encoding error
    ),
)
def test_string_scannable_is_longer_than(content, is_longer):
    """
    GIVEN a StringScannable
    WHEN is_longer_than() is called
    THEN it returns the expected value
    """
    scannable = StringScannable(content=content, url="u")
    assert scannable.is_longer_than(50) == is_longer



from copy import deepcopy
from operator import itemgetter
from pathlib import Path
from typing import Any, Dict

import yaml

from ggshield.__main__ import cli
from tests.unit.conftest import assert_invoke_ok


V1_CONFIG_CONTENT = """
paths-ignore:
  - '**/migrations/**/*'
  - '**/snap*'
  - '.gitlab/*'
  - 'LICENSE'

matches-ignore:
  - name:
    match: vLXyx1iAhFo2xgb71tTa
  - name: generic password
    match: 05Panda_8463
  - name: test github oauth password
    match: 29825c15f543be6158140d0561a2257a5157ee6d845355c62eb1b53bfd4155af
  - v1.6793be7966338465559c751561e82de746880ccb

all-policies: true
show-secrets: true
"""

V2_CONFIG_DICT = {
    "version": 2,
    "secret": {
        "ignored_paths": [
            "**/snap*",
            "**/migrations/**/*",
            ".gitlab/*",
            "LICENSE",
        ],
        "show_secrets": True,
        "ignored_matches": [
            {"name": "", "match": "vLXyx1iAhFo2xgb71tTa"},
            {"name": "generic password", "match": "05Panda_8463"},
            {
                "name": "test github oauth password",
                "match": "29825c15f543be6158140d0561a2257a5157ee6d845355c62eb1b53bfd4155af",
            },
            {"name": "", "match": "v1.6793be7966338465559c751561e82de746880ccb"},
        ],
    },
}


def normalize_config_dict(dct: Dict[str, Any]) -> Dict[str, Any]:
    """
    Sort lists inside a config dict so that the dicts can be compared with ==
    """
    dct = deepcopy(dct)
    try:
        dct["secret"]["ignored_paths"] = sorted(dct["secret"]["ignored_paths"])
    except KeyError:
        pass
    try:
        dct["secret"]["ignored_matches"] = sorted(
            dct["secret"]["ignored_matches"], key=itemgetter("match")
        )
    except KeyError:
        pass
    return dct


def test_config_migrate_cmd(cli_fs_runner):
    """
    GIVEN a v1 config file
    WHEN `ggshield config migrate` is called
    THEN the file is migrated to v2
    AND the v1 config file is kept as a backup
    """
    config_path = Path(".gitguardian.yaml")
    config_path.write_text(V1_CONFIG_CONTENT)

    result = cli_fs_runner.invoke(cli, ["config", "migrate"])
    assert_invoke_ok(result)

    # Check new file is v2
    with config_path.open() as f:
        dct = yaml.safe_load(f)
    assert normalize_config_dict(dct) == normalize_config_dict(V2_CONFIG_DICT)

    # Check backup is unchanged
    assert ".gitguardian.yaml.old" in result.stdout
    assert Path(".gitguardian.yaml.old").read_text() == V1_CONFIG_CONTENT



import json
import os

import pytest
import yaml

from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.types import IgnoredMatch


@pytest.mark.usefixtures("isolated_fs")
class TestCache:
    def test_defaults(self):
        cache = Cache()
        assert cache.last_found_secrets == []

    def test_load_cache_and_purge(self):
        with open(".cache_ggshield", "w") as file:
            json.dump({"last_found_secrets": [{"name": "", "match": "XXX"}]}, file)
        cache = Cache()
        assert cache.last_found_secrets == [IgnoredMatch(name="", match="XXX")]

        cache.purge()
        assert cache.last_found_secrets == []

    def test_load_invalid_cache(self, capsys):
        with open(".cache_ggshield", "w") as file:
            json.dump({"invalid_option": True}, file)

        Cache()
        captured = capsys.readouterr()
        assert "Unrecognized key in cache" in captured.err

    def test_save_cache(self):
        with open(".cache_ggshield", "w") as file:
            json.dump({}, file)
        cache = Cache()
        cache.update_cache(last_found_secrets=[{"match": "XXX"}])
        cache.save()
        with open(".cache_ggshield") as file:
            file_content = json.load(file)
            assert file_content == {
                "last_found_secrets": [{"match": "XXX", "name": ""}]
            }

    def test_read_only_fs(self):
        """
        GIVEN a read-only cache file
        WHEN save is called
        THEN it shouldn't raise an exception
        """

        cache = Cache()
        cache.update_cache(last_found_secrets=[{"match": "XXX"}])

        # Make cache file read-only and verify it really is read-only
        cache.cache_path.touch(mode=0o400)
        with pytest.raises(PermissionError):
            cache.cache_path.open("w")

        cache.save()

    @pytest.mark.parametrize("with_entry", [True, False])
    def test_save_cache_first_time(self, isolated_fs, with_entry):
        """
        GIVEN no existing cache
        WHEN save is called but there are (new entries/no entries in memory)
        THEN it should (create/not create) the file
        """
        cache = Cache()
        if with_entry:
            cache.update_cache(last_found_secrets=[{"match": "XXX"}])
        cache.save()

        assert os.path.isfile(".cache_ggshield") is with_entry

    def test_max_commits_for_hook_setting(self):
        """
        GIVEN a yaml config with `max-commits-for-hook=75`
        WHEN the config gets parsed
        THEN the default value of max_commits_for_hook (50) should be replaced with 75
        """
        with open(".gitguardian.yml", "w") as file:
            file.write(yaml.dump({"max-commits-for-hook": 75}))

        config = Config()
        assert config.user_config.max_commits_for_hook == 75



from ggshield.core.text_utils import format_text


def format_quota_color(remaining: int, limit: int) -> str:
    if limit == 0:
        return format_text(str(remaining), {"fg": "white"})

    available_percent = remaining / limit
    if available_percent < 0.25:
        color = "red"
    elif available_percent < 0.75:
        color = "yellow"
    else:
        color = "green"

    return format_text(str(remaining), {"fg": color})



import codecs
from pathlib import Path
from random import randrange

import charset_normalizer
import pytest

from ggshield.core.scan import DecodeError, File
from ggshield.core.scan.file import create_files_from_paths
from tests.conftest import is_windows


UNICODE_TEST_CONTENT = "Ascii 123, accents: éèà, hiragana: ぁ, emoji: 🛡️"


def get_charset_normalizer_encoding(path: Path) -> str:
    """Returns the encoding that charset_normalizer would detect for `path`"""
    charset_match = charset_normalizer.from_path(path)
    assert charset_match
    return charset_match.best().encoding


@pytest.mark.parametrize(
    ["encoding", "bom"],
    [
        ("utf-8", b""),
        ("utf-8", codecs.BOM_UTF8),
        ("utf-16-be", b""),
        ("utf-16-be", codecs.BOM_UTF16_BE),
        ("utf-16-le", b""),
        ("utf-16-le", codecs.BOM_UTF16_LE),
        ("utf-32-be", b""),
        ("utf-32-be", codecs.BOM_UTF32_BE),
        ("utf-32-le", b""),
        ("utf-32-le", codecs.BOM_UTF32_LE),
    ],
)
def test_file_decode_content(tmp_path, encoding: str, bom: bytes):
    """
    GIVEN a valid utf encoded file, with or without a BOM
    WHEN FILE tries to decode it
    THEN it succeeds
    """
    path = tmp_path / "test.conf"
    raw_content = bom + UNICODE_TEST_CONTENT.encode(encoding)
    path.write_bytes(raw_content)
    file = File(path)
    assert file.content == UNICODE_TEST_CONTENT


def test_file_does_not_decode_binary(tmp_path):
    """
    GIVEN a 2000 random bytes file
    WHEN File tries to decode it
    THEN it raises a DecodeError
    """
    path = tmp_path / "test.conf"
    data = (randrange(256) for _ in range(2000))
    path.write_bytes(bytes(data))

    file = File(path)
    with pytest.raises(DecodeError):
        _dummy = file.content  # noqa (mute "_dummy" is never used)


def test_file_is_longer_does_not_decode_binary(tmp_path):
    """
    GIVEN a 2000 random bytes file
    WHEN is_longer_than() is called on it
    THEN it raises a DecodeError
    """
    path = tmp_path / "test.conf"
    data = (randrange(256) for _ in range(2000))
    path.write_bytes(bytes(data))

    file = File(path)
    with pytest.raises(DecodeError):
        file.is_longer_than(1000)


def test_file_is_longer_than_does_not_read_large_files(tmp_path):
    """
    GIVEN a File instance on a large file
    WHEN is_longer_than() is called on it
    THEN it returns the right value
    AND the content is not read
    """
    path = tmp_path / "test.conf"
    path.write_text(UNICODE_TEST_CONTENT * 1000, encoding="utf-8")

    file = File(path)
    assert file.is_longer_than(50)
    assert file._content is None


def test_file_is_longer_than_does_not_read_utf8_file(tmp_path):
    """
    GIVEN a File instance on either a small utf8 file
    WHEN is_longer_than() is called on it
    THEN it returns False
    AND the content is not read
    """
    path = tmp_path / "test.conf"
    path.write_text(UNICODE_TEST_CONTENT, encoding="utf-8")
    assert get_charset_normalizer_encoding(path) == "utf_8"

    file = File(path)
    assert not file.is_longer_than(1000)
    assert file._content is None


def test_file_is_longer_than_if_file_has_been_read(tmp_path):
    """
    GIVEN a File instance
    AND it has already read its file
    WHEN is_longer_than() is called on it
    THEN it returns the right value
    """
    path = tmp_path / "test.conf"
    content = "Mangé"
    path.write_text(content, encoding="utf-8")

    file = File(path)
    byte_content = path.stat().st_size

    # byte_content should be greater than len(content) because the *utf-8 encoded*
    # content is longer than the str content
    assert byte_content > len(content)

    # Force reading
    assert file.content == content

    assert file.is_longer_than(len(content))

    assert not file.is_longer_than(byte_content)


def test_file_is_longer_utf32(tmp_path):
    """
    GIVEN a file encoded in utf32, whose byte size is greater than N but whose utf8
    size is less than N
    WHEN is_longer_than(N) is called on it
    THEN it returns False
    AND the content is available because is_longer_than() read all the file
    """
    path = tmp_path / "test.conf"
    str_size = 200
    str_content = "x" * str_size

    byte_content = str_content.encode("utf32")
    path.write_bytes(byte_content)
    assert get_charset_normalizer_encoding(path) == "utf_32"

    # byte_content is longer than str_size because utf32 uses 4 bytes per code-point
    assert len(byte_content) > str_size

    file = File(path)
    assert not file.is_longer_than(len(byte_content))
    assert file._content == str_content


def test_file_is_longer_using_8bit_codec(tmp_path):
    """
    GIVEN a file encoded using an 8bit codec, whose byte size is less than N but whose
    utf8 size is greater than N
    WHEN is_longer_than(N) is called on it
    THEN it returns True
    """
    path = tmp_path / "test.conf"
    # Use characters that require more than one byte in utf8
    str_content = "écrit en français€"

    byte_content = str_content.encode("cp1250")
    path.write_bytes(byte_content)
    assert get_charset_normalizer_encoding(path) == "cp1250"

    # byte_content is shorter than str_content encoded as utf8 because cp1250 uses only
    # 1 byte per character
    assert len(byte_content) < len(str_content.encode())

    file = File(path)
    assert file.is_longer_than(len(byte_content))


def test_file_repr():
    """
    GIVEN a File instance
    WHEN repr() is called
    THEN it returns the correct output
    """
    if is_windows():
        str_path = r"c:\Windows"
        expected_url = "file:///c:/Windows"
    else:
        str_path = "/usr"
        expected_url = "file:///usr"
    file = File(Path(str_path))
    assert repr(file) == f"<File url={expected_url} filemode=Filemode.FILE>"


def test_file_path():
    """
    GIVEN an OS-specific path
    WHEN creating a File on it
    THEN file.path returns the correct path
    """
    str_path = r"c:\Windows" if is_windows() else "/usr"
    file = File(Path(str_path))
    assert file.path == Path(str_path)


@pytest.mark.parametrize(
    ["filename", "input_content", "expected_content"],
    [
        ("normal.txt", b"Normal", "Normal"),
        ("invalid-utf8-start-byte.txt", b"Hello\x81World", "Hello·World"),
        ("zero-bytes-are-kept.txt", b"Zero\0byte", "Zero\0byte"),
    ],
)
def test_create_files_from_paths(
    tmp_path, filename: str, input_content: bytes, expected_content: str
):
    """
    GIVEN a file
    WHEN calling create_files_from_paths() on it
    THEN it returns the expected File instance
    AND the content of the File instance is what is expected
    """
    path = tmp_path / filename
    path.write_bytes(input_content)

    files, _ = create_files_from_paths([path], exclusion_regexes=set())

    file = files[0]
    assert file.filename == str(path)
    assert file.content == expected_content

    assert len(files) == 1



import os
import platform
import tarfile
from io import BytesIO
from pathlib import Path
from typing import Optional
from unittest.mock import Mock, patch

import pytest

from ggshield.core.tar_utils import tar_from_ref_and_filepaths
from ggshield.utils.git_shell import (
    InvalidGitRefError,
    NotAGitDirectory,
    check_git_dir,
    check_git_ref,
    get_default_branch,
    get_filepaths_from_ref,
    get_new_branch_ci_commits,
    get_remotes,
    get_repository_url_from_path,
    get_staged_filepaths,
    git,
    git_ls_unstaged,
    is_git_dir,
    is_valid_git_commit_ref,
    simplify_git_url,
)
from ggshield.utils.os import cd
from tests.repository import Repository


def _add_remote(
    repository: Repository, repository_name: str, remote_name: Optional[str] = "origin"
):
    remote_url = f"https://github.com/owner/{repository_name}.git"
    repository.git("remote", "add", remote_name, remote_url)


def _create_repository_with_remote(
    repository_path: Path,
    repository_name: str,
    remote_name: Optional[str] = "origin",
):
    local_repo = Repository.create(repository_path, bare=True)
    _add_remote(local_repo, repository_name, remote_name)
    return local_repo


def test_git_shell():
    assert "usage: git" in git(["help"])


def test_is_git_dir(tmp_path):
    assert is_git_dir(os.getcwd())
    assert not is_git_dir(str(tmp_path))


def test_is_valid_git_commit_ref():
    assert is_valid_git_commit_ref("HEAD")
    assert not is_valid_git_commit_ref("invalid_ref")


def test_get_new_branch_ci_commits(tmp_path: Path):
    # GIVEN a remote repository
    remote_repository = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone, with commitM pushed on main
    local_repository = Repository.clone(remote_repository.path, tmp_path / "local")
    local_repository.create_commit()
    local_repository.push()

    # AND commitA1 on new branchA
    local_repository.create_branch("branchA")
    commitA1_sha = local_repository.create_commit("commitA1")
    # AND commitB1 & B2 on new branchB, created from branchA
    local_repository.create_branch("branchB")
    commitB1_sha = local_repository.create_commit("commitB1")
    commitB2_sha = local_repository.create_commit("commitB2")
    # AND commitA2 created later on branchA
    local_repository.checkout("branchA")
    commitA2_sha = local_repository.create_commit("commitA2")

    # WHEN executing get_branch_new_commits
    # THEN only new commits are found for each branch
    local_repository.checkout("branchA")
    assert get_new_branch_ci_commits("branchA", local_repository.path) == [
        commitA2_sha,
        commitA1_sha,
    ]
    local_repository.checkout("branchB")
    assert get_new_branch_ci_commits("branchB", local_repository.path) == (
        [commitB2_sha, commitB1_sha, commitA1_sha]
    )


def test_get_branch_new_commits_many_branches(tmp_path: Path):
    # GIVEN a repository with many branches
    remote_repository = Repository.create(tmp_path / "remote", bare=True)
    local_repository = Repository.clone(remote_repository.path, tmp_path / "local")
    for i in range(50):
        local_repository.create_branch(f"branch_with_a_long_name{i}")
        local_repository.create_commit()
    local_repository.push("--all")

    # AND a local branch with one new commit
    local_repository.create_branch("tested-branch")
    tested_commit = local_repository.create_commit()
    # WHEN executing get_branch_new_commits
    # THEN it works as intended
    assert get_new_branch_ci_commits("tested-branch", local_repository.path) == [
        tested_commit
    ]


def test_check_git_dir(tmp_path):
    """
    GIVEN a git checkout
    AND check_git_dir() has been called without arguments in it
    AND it did not raise an exception
    WHEN the current directory is changed to a directory which is not a git checkout
    AND check_git_dir() is called without arguments
    THEN it raises an exception

    (this tests the LRU cache on the functions in git_shell.py works correctly)
    """
    check_git_dir()

    with cd(str(tmp_path)):
        with pytest.raises(NotAGitDirectory):
            check_git_dir()


def test_check_git_ref_invalid_git_path(tmp_path):
    # WHEN checking a non git path
    with cd(str(tmp_path)):
        # THEN function throws an error
        with pytest.raises(NotAGitDirectory):
            check_git_ref(ref="HEAD")


def test_check_git_ref_valid_git_path(tmp_path):
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo_path = tmp_path / "local"
    local_repo = Repository.clone(remote_repo.path, local_repo_path)
    local_repo.create_commit()
    local_repo.push()

    # THEN valid git references do not throw
    check_git_ref("HEAD", local_repo_path)
    check_git_ref("@{upstream}", local_repo_path)

    # AND other strings throw
    with pytest.raises(InvalidGitRefError):
        check_git_ref("invalid_ref", local_repo_path)


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "https://user:password@github.com:84/GitGuardian/ggshield.git",
            "github.com/GitGuardian/ggshield",
        ),
        (
            "https://github.com/GitGuardian/ggshield.git",
            "github.com/GitGuardian/ggshield",
        ),
        (
            "git@github.com:GitGuardian/ggshield.git",
            "github.com/GitGuardian/ggshield",
        ),
        (
            "https://github.com/Git.Guar-di_an/gg.sh-ie_ld.git",
            "github.com/Git.Guar-di_an/gg.sh-ie_ld",
        ),
        (
            "https://gitlab.instance.ovh/owner/project/repository.git",
            "gitlab.instance.ovh/owner/project/repository",
        ),
        (
            "https://username@dev.azure.com/username/project/_git/repository",
            "dev.azure.com/username/project/repository",
        ),
        (
            "https://username@bitbucket.org/owner/repository.git",
            "bitbucket.org/owner/repository",
        ),
    ],
    ids=[
        "Full Github https",
        "Github https",
        "Github ssh",
        "Github special characters",
        "Gitlab https",
        "Azure Devops https",
        "BitBucket https",
    ],
)
def test_simplify_git_url(url, expected):
    assert expected == simplify_git_url(url)


def test_get_repository_url_from_path(tmp_path: Path):
    # GIVEN a local repository with remote url
    local_repo = _create_repository_with_remote(tmp_path, "repository")

    # THEN the remote url is returned in the root clone directory
    assert "repository" in get_repository_url_from_path(local_repo.path)
    # AND in a subdirectory
    subdirectory_path = local_repo.path / "subdirectory"
    subdirectory_path.mkdir()
    assert "repository" in get_repository_url_from_path(subdirectory_path)


def test_get_repository_url_from_path_no_repo(tmp_path: Path):
    # GIVEN a local directory with no remote git directory
    local_directory_path = tmp_path / "local"
    local_directory_path.mkdir()
    # AND a local repository with no remote git directory
    local_repository_path = tmp_path / "repo"
    repo = Repository.create(local_repository_path)
    repo.create_commit()

    # THEN no url is returned
    assert get_repository_url_from_path(local_directory_path) is None
    assert get_repository_url_from_path(local_repository_path) is None


def test_get_repository_url_from_path_two_remotes(tmp_path: Path):
    # GIVEN a local repository with two remotes
    local_repo = _create_repository_with_remote(
        repository_path=tmp_path,
        repository_name="repository1",
        remote_name="other_remote",
    )
    _add_remote(
        repository=local_repo,
        repository_name="repository2",
        remote_name="origin",
    )

    # THEN only one remote is returned, with priority to origin
    assert "repository2" in get_repository_url_from_path(local_repo.path)
    # AND in a subdirectory
    subdirectory_path = local_repo.path / "subdirectory"
    subdirectory_path.mkdir()
    assert "repository2" in get_repository_url_from_path(subdirectory_path)


def test_get_repository_url_from_path_different_repo(tmp_path: Path):
    # GIVEN two repositories with one remote each
    local_repo1 = _create_repository_with_remote(
        repository_path=tmp_path / "local1",
        repository_name="repository1",
    )
    local_repo2 = _create_repository_with_remote(
        repository_path=tmp_path / "local2",
        repository_name="repository2",
    )

    # THEN scanning repo 2 from repo 1 yields repo 2's remote url
    with cd(str(local_repo1.path)):
        assert "repository2" in get_repository_url_from_path(local_repo2.path)


def test_get_repository_url_from_path_subrepo(tmp_path: Path):
    # GIVEN two repositories, each with its remote, with repo2 nested inside repo1
    local_repo1 = _create_repository_with_remote(
        repository_path=tmp_path,
        repository_name="repository1",
    )
    local_repo2 = _create_repository_with_remote(
        repository_path=local_repo1.path / "nested",
        repository_name="repository2",
    )

    # THEN scanning local repo 1 returns remote repo 1 url
    assert "repository1" in get_repository_url_from_path(local_repo1.path)
    # AND scanning local repo 2 returns remote repo 2 url
    assert "repository2" in get_repository_url_from_path(local_repo2.path)


def test_get_filepaths_from_ref(tmp_path):
    # GIVEN a repository
    repo = Repository.create(tmp_path)
    repo.create_commit()

    # AND a first commit
    first_file = repo.path / "first.py"
    first_content = "First file (included)"
    first_file.write_text(first_content)
    repo.add("first.py")
    repo.create_commit()

    # AND a second commit
    second_file = repo.path / "second.py"
    second_content = "Second file (not included)"
    second_file.write_text(second_content)
    repo.add("second.py")
    repo.create_commit()

    # WHEN scanning since the second commit
    filepaths = [str(path) for path in get_filepaths_from_ref("HEAD~1", tmp_path)]

    # THEN file from first commit is part of filepaths
    assert "first.py" in filepaths
    # AND file from second commit is not part of filepaths
    assert "second.py" not in filepaths


def test_get_staged_filepaths(tmp_path):
    # GIVEN a repository
    repo = Repository.create(tmp_path)
    repo.create_commit()

    # AND a first commit
    first_file = repo.path / "first.py"
    first_content = "First file (included)"
    first_file.write_text(first_content)
    repo.add("first.py")
    repo.create_commit()

    # AND staged content
    second_file = repo.path / "second.py"
    second_content = "Second file (included)"
    second_file.write_text(second_content)
    repo.add("second.py")

    # WHEN scanning for files, including staged
    filepaths = [str(path) for path in get_staged_filepaths(tmp_path)]

    # THEN file from first commit is part of filepaths
    assert "first.py" in filepaths
    # AND staged file is part of filepaths
    assert "second.py" in filepaths


def test_tar_from_ref_and_filepaths(tmp_path):
    # GIVEN a repository
    repo = Repository.create(tmp_path)
    repo.create_commit()

    first_file_name = "first.py"
    first_ignored_file_name = "first_ignored.py"
    second_file_name = "second.py"

    # AND a first commit
    first_file = repo.path / first_file_name
    first_content = "First file (included)"
    first_file.write_text(first_content)
    repo.add(first_file_name)

    first_ignored_file = repo.path / first_ignored_file_name
    first_ignored_content = "First file (filtered out)"
    first_ignored_file.write_text(first_ignored_content)
    repo.add(first_ignored_file_name)
    repo.create_commit()

    # AND a second commit
    second_file = repo.path / second_file_name
    second_content = "Second file (not included)"
    second_file.write_text(second_content)
    repo.add(second_file_name)
    repo.create_commit()

    # AND a list of filepaths
    filepaths = [first_file_name]

    # WHEN creating a tar
    tarbytes = tar_from_ref_and_filepaths(
        "HEAD~1", [Path(path_str) for path_str in filepaths], wd=tmp_path
    )

    tar_stream = BytesIO(tarbytes)
    with tarfile.open(fileobj=tar_stream, mode="r:gz") as tar:
        filenames = tar.getnames()
        # THEN only first file in in tar
        assert filenames == [first_file_name]


def test_get_remotes(tmp_path):
    # GIVEN a repository with two remotes
    repo = Repository.create(tmp_path)
    repo.git("remote", "add", "origin", "https://a_remote_repo")
    repo.git("remote", "add", "upstream", "https://another_remote_repo")

    # wHEN listing remotes
    remotes = get_remotes(repo.path)

    # THEN all the remotes are listed
    assert set(remotes) == {"origin", "upstream"}


def test_get_default_branch_with_remote(tmp_path):
    # GIVEN a repository with a remote
    remote_repository = Repository.create(tmp_path / "remote", initial_branch="foo")
    remote_repository.create_commit()
    local_repository = Repository.clone(remote_repository.path, tmp_path / "local")

    # WHEN getting the default branch
    default_branch = get_default_branch(local_repository.path)

    # THEN the default branch is the remote initial branch
    assert default_branch == "origin/foo"


def test_get_default_branch_without_remote(tmp_path):
    # GIVEN a repository without a remote
    local_repository = Repository.create(tmp_path / "remote", initial_branch="foo")
    local_repository.git("config", "--local", "init.defaultBranch", "foo")

    # WHEN getting the default branch
    default_branch = get_default_branch(local_repository.path)

    # THEN the default branch is the initial branch
    assert default_branch == "foo"


def test_git_ls_unstaged(tmp_path):
    # GIVEN a repository
    repo = Repository.create(tmp_path)
    repo.create_commit()

    # AND a submodule
    submodule = Repository.create(repo.path / "submodule")
    submodule.create_commit()

    repo.git("submodule", "add", submodule.path)

    # AND unstaged files
    repo_file = repo.path / "repo_file"
    submodule_file = repo.path / "submodule" / "submodule_file"
    repo_file.write_text("Content")
    submodule_file.write_text("Content")

    # WHEN listing unstaged files
    unstaged_files = git_ls_unstaged(wd=repo.path)

    # THEN unstaged files in repo and submodule are returned
    # as relative to repo.path
    expected_paths = {x.relative_to(repo.path) for x in (repo_file, submodule_file)}
    assert {Path(x) for x in unstaged_files} == expected_paths


@patch("subprocess.run", return_value=Mock(stdout=b""))
def test_git_command_includes_longpaths_on_windows(mock_run):
    # GIVEN any git command
    git(["status"])
    mock_run.assert_called_once()

    # THEN the command includes core.longpaths=true if on Windows
    command = mock_run.call_args[0][0]
    longpaths_included = any(param == "core.longpaths=true" for param in command)
    if platform.system() == "Windows":
        assert (
            longpaths_included
        ), f"core.longpaths=true not found in command: {command}"
    else:
        assert (
            not longpaths_included
        ), f"core.longpaths=true found in command: {command}"



from datetime import datetime, timezone
from typing import Optional

import pytest

from ggshield.core.config import Config
from ggshield.core.config.user_config import CURRENT_CONFIG_VERSION, UserConfig
from ggshield.core.errors import ParseError, UnexpectedError
from ggshield.core.types import IgnoredMatch
from tests.unit.conftest import write_text, write_yaml


@pytest.mark.usefixtures("isolated_fs")
class TestUserConfig:
    def _assert_times(self, time1: Optional[datetime], time2: Optional[str]):
        if time1 is not None:
            assert time1 == datetime.strptime(
                str(time2), "%Y-%m-%d %H:%M:%S"
            ).astimezone(timezone.utc)

    def test_parsing_error(cli_fs_runner, local_config_path):
        write_text(local_config_path, "Not a:\nyaml file.\n")
        expected_output = f"{local_config_path} is not a valid YAML file:"
        with pytest.raises(ParseError, match=expected_output):
            Config()

    def test_display_options(self, local_config_path):
        write_yaml(local_config_path, {"verbose": True, "show_secrets": True})

        config = Config()
        assert config.user_config.verbose is True
        assert config.user_config.secret.show_secrets is True

    def test_display_options_inheritance(self, local_config_path, global_config_path):
        write_yaml(
            local_config_path,
            {
                "verbose": True,
                "show_secrets": False,
                "api_url": "https://api.gitguardian.com",
            },
        )
        write_yaml(
            global_config_path,
            {
                "verbose": False,
                "show_secrets": True,
                "api_url": "https://api.gitguardian.com2",
            },
        )

        config = Config()
        assert config.user_config.verbose is True
        assert config.user_config.secret.show_secrets is False
        assert config.api_url == "https://api.gitguardian.com"

    def test_exclude_regex(self, local_config_path):
        write_yaml(local_config_path, {"paths-ignore": ["/tests/"]})

        config = Config()
        assert r"/tests/" in config.user_config.secret.ignored_paths

    def test_accumulation_matches(self, local_config_path, global_config_path):
        write_yaml(
            local_config_path,
            {
                "matches_ignore": [
                    {"name": "", "match": "one"},
                    {"name": "", "match": "two"},
                ]
            },
        )
        write_yaml(
            global_config_path,
            {"matches_ignore": [{"name": "", "match": "three"}]},
        )
        config = Config()
        assert config.user_config.secret.ignored_matches == [
            IgnoredMatch(match="three", name=""),
            IgnoredMatch(match="one", name=""),
            IgnoredMatch(match="two", name=""),
        ]

    def test_load_too_new_version(self, local_config_path):
        """
        GIVEN a config file whose format is too recent
        WHEN we try to load it
        THEN an exception is raised
        """
        write_yaml(local_config_path, {"version": CURRENT_CONFIG_VERSION + 1})

        with pytest.raises(UnexpectedError):
            UserConfig.load(local_config_path)

    def test_save_load_roundtrip(self):
        config_path = "config.yaml"
        config = UserConfig()
        config.exit_zero = True
        config.secret.show_secrets = True

        config.save(config_path)

        config2, _ = UserConfig.load(config_path)

        assert config.exit_zero == config2.exit_zero
        assert config.secret.show_secrets == config2.secret.show_secrets

    def test_load_v1(self):
        config_path = "config.yaml"
        write_yaml(
            config_path,
            {
                "exit-zero": True,
                "show-secrets": True,
                "banlisted-detectors": ["d1", "d2"],
                "matches-ignore": [
                    {
                        "name": "foo",
                        "match": "abcdef",
                    },
                    # A match using the old format: just a sha256
                    "1234abcd",
                ],
                "paths-ignore": ["/foo", "/bar"],
            },
        )

        config, _ = UserConfig.load(config_path)

        assert config.exit_zero
        assert config.secret.show_secrets
        assert config.secret.ignored_detectors == {"d1", "d2"}
        assert config.secret.ignored_matches == [
            IgnoredMatch(name="foo", match="abcdef"),
            IgnoredMatch(name="", match="1234abcd"),
        ]
        assert config.secret.ignored_paths == {"/foo", "/bar"}

    def test_load_mixed_versions(self, local_config_path, global_config_path):
        write_yaml(global_config_path, {"paths-ignore": ["v1path"]})

        write_yaml(
            local_config_path, {"version": 2, "secret": {"ignored_paths": ["v2path"]}}
        )

        config, _ = UserConfig.load()
        assert config.secret.ignored_paths == {"v1path", "v2path"}

    def test_load_ignored_matches_with_empty_names(self):
        config_path = "config.yaml"
        # Use write_text() here because write_yaml() cannot generate a key with a really
        # empty value, like we need for secret.ignored_matches[0].name
        write_text(
            config_path,
            """
            version: 2
            secret:
              ignored_matches:
                - name:
                  match: abcd
                - name: ""
                  match: dbca
            """,
        )
        config, _ = UserConfig.load(config_path)

        assert config.secret.ignored_matches == [
            IgnoredMatch(name="", match="abcd"),
            IgnoredMatch(name="", match="dbca"),
        ]

    def test_config_options_inheritance(self, local_config_path, global_config_path):
        """
        GIVEN two config files (global and local)
        WHEN deserializing them
        THEN the inheritance is respected
        """
        write_yaml(
            global_config_path,
            {
                "version": 2,
                "secret": {
                    "show_secrets": True,
                    "ignored_paths": ["myglobalpath"],
                },
            },
        )
        write_yaml(
            local_config_path,
            {
                "version": 2,
                "secret": {
                    "ignored_paths": ["mypath"],
                },
            },
        )
        config = Config()

        # Checks the global config value for `show_secrets` is still there after loading
        # the local config, which does not contain this key.
        assert config.user_config.secret.show_secrets

        secret_config = config.user_config.secret
        assert secret_config.ignored_paths == {"myglobalpath", "mypath"}

    def test_user_config_unknown_keys(self, local_config_path, capsys):
        """
        GIVEN a config containing unknown keys
        WHEN deserializing it
        THEN the keys are ignored and a warning is raised for config keys
        """
        write_yaml(
            local_config_path,
            {
                "version": 2,
                "root_unknown": "false key",
                "secret": {
                    "secret_invalid_key": "invalid key",
                    "ignored_matches": [
                        {
                            "name": "",
                            "match": "one",
                            "match_invalid_key": "two",
                            "nested-hashed": "example",
                        },
                    ],
                },
                "hashed-key": "example",
            },
        )
        UserConfig.load(local_config_path)
        captured = capsys.readouterr()
        assert "Unrecognized key in config: root_unknown" in captured.err
        assert "Unrecognized key in config: secret_invalid_key" in captured.err
        assert "Unrecognized key in config: match_invalid_key" in captured.err
        assert "Unrecognized key in config: hashed_key" in captured.err
        assert "Unrecognized key in config: nested_hashed" in captured.err

    def test_user_config_dash_keys(self, local_config_path, capsys):
        """
        GIVEN a config containing keys separated with dashes
        WHEN deserializing it
        THEN the keys are loaded and a warning is raised for config keys
        """
        write_yaml(
            local_config_path,
            {
                "version": 2,
                "secret": {
                    "ignored-paths": ["myglobalpath"],
                },
            },
        )
        cfg, _ = UserConfig.load(local_config_path)
        captured = capsys.readouterr()

        assert cfg.secret.ignored_paths == {"myglobalpath"}
        assert (
            f"{local_config_path}: Config key ignored-paths is deprecated, use ignored_paths instead."
            in captured.err
        )

    def test_can_load_ignored_known_secrets_from_root(self, local_config_path):
        """
        GIVEN a config file containing the `ignore_known_secrets` key in the root mapping
        WHEN deserializing it
        THEN UserConfig.secret.ignore_known_secrets has the correct value
        """
        write_yaml(
            local_config_path,
            {
                "version": 2,
                "ignore_known_secrets": True,
            },
        )
        config, _ = UserConfig.load(local_config_path)
        assert config.secret.ignore_known_secrets

    def test_can_load_ignored_known_secrets_from_secret(self, local_config_path):
        """
        GIVEN a config file containing the `ignore_known_secrets` key in the `secret` mapping
        WHEN deserializing it
        THEN UserConfig.secret.ignore_known_secrets has the correct value
        """
        write_yaml(
            local_config_path,
            {
                "version": 2,
                "secret": {
                    "ignore-known-secrets": True,
                },
            },
        )
        config, _ = UserConfig.load(local_config_path)
        assert config.secret.ignore_known_secrets

    def test_bad_local_config(self, local_config_path, global_config_path):
        """
        GIVEN a malformed .gitguardian.yaml, with a list of instance
        WHEN loading the user condiguration
        THEN an error is returned
        """
        write_yaml(global_config_path, {"instance": "https://test.gitguardian.com/"})
        write_yaml(
            local_config_path, {"instance": ["https://dashboard.gitguardian.com/"]}
        )
        with pytest.raises(UnexpectedError):
            UserConfig.load()



"""
Main entry point to produce progress output.

Keeps a GGShieldUI instance at module level and exposes the instance output methods as
plain functions.
"""

from logging import LogRecord

from .ggshield_ui import GGShieldProgress, GGShieldUI, Level
from .plain_text import PlainTextGGShieldUI


# GGShieldUI instance to which top-level functions forward their output.
# Can be changed using set_ui().
_ui: GGShieldUI = PlainTextGGShieldUI()


def set_level(level: Level) -> None:
    _ui.level = level


def get_level() -> Level:
    return _ui.level


def ensure_level(level: Level):
    """
    Make sure the verbosity level is at least set to `level`
    """
    if _ui.level < level:
        set_level(level)


def is_verbose() -> bool:
    """
    Convenient function to check if verbose messages are visible. Use this if displaying
    verbose messages is costly (for example displaying a list of files)
    """
    return _ui.level >= Level.VERBOSE


def set_ui(ui: GGShieldUI) -> None:
    """Change the GGShieldUI instance used to output messages. Takes care of
    carrying existing settings from the old instance to the new one."""
    global _ui
    ui.level = _ui.level
    _ui = ui


def get_ui() -> GGShieldUI:
    return _ui


def display_debug(message: str) -> None:
    _ui.display_debug(message)


def display_verbose(message: str) -> None:
    _ui.display_verbose(message)


def display_info(message: str) -> None:
    _ui.display_info(message)


def display_heading(message: str) -> None:
    _ui.display_heading(message)


def display_warning(message: str) -> None:
    _ui.display_warning(message)


def display_error(message: str) -> None:
    _ui.display_error(message)


def log(record: LogRecord) -> None:
    _ui.log(record)


def create_progress(total: int) -> GGShieldProgress:
    return _ui.create_progress(total)


def _reset_ui():
    """Reset the module to its startup state. Used by reset.reset()."""
    global _ui
    _ui = PlainTextGGShieldUI()
